# basics
import argparse
from argparse import ArgumentParser
import os
import os.path as op
import time
import json
import pandas as pd
import numpy as np
import scipy

# ML pipeline
import datasets
import torch
import matplotlib.pyplot as plt
from torch.nn import CrossEntropyLoss
from torch.utils.data import DataLoader
from sklearn import metrics

from datasets import load_dataset, DatasetDict, Dataset, load_from_disk
from transformers import TrainingArguments, Trainer, AutoTokenizer
from transformers import GPTNeoXForCausalLM, GPTNeoXForSequenceClassification # PYTHIA
from transformers import AutoModelForCausalLM, AutoModelForSequenceClassification # BLOOM
from transformers import GPT2Model, GPT2ForSequenceClassification # GPT2

from ml_collections import config_dict


def eval_accuracy(model, data_loader, device: str):
    model.to(device).eval()
    acc = 0
    for idx, batch in enumerate(data_loader):
        # get input embeddings
        inputs_embeds = batch['input_embeds'].to(device)
        attention_mask = batch['attention_mask'].to(device)
        labels = batch['label'].to(device).clone().detach().cpu().numpy()
        batch_size = len(labels)
        # compute loss
        pred_scores = model(inputs_embeds=inputs_embeds, attention_mask=attention_mask).logits.clone().detach().cpu().numpy()
        preds = pred_scores.argmax(1)
        acc += ((preds == labels)*1).mean()
    acc = acc / (idx +1)
    return acc

def compute_curve(train_scores: np.array, test_scores: np.array, pos_label: int=1):
    y = np.r_[np.ones(np.shape(train_scores)[0]), np.zeros(np.shape(test_scores)[0])]
    fs, ts, thresholds = metrics.roc_curve(y, np.r_[-np.array(train_scores), -np.array(test_scores)], pos_label=pos_label)
    return ts, fs

def add_suffix(sample):
    if sample['label'] == 0:
        label = 'negative'
    else:
        label = 'positive'
    sample['text'] = sample['text'] + ' // ' + label
    return sample
    
def add_suffix_ag(sample):
    if sample['label'] == 0:
        label = 'world'
    elif sample['label'] == 1:
        label = 'sports'
    elif sample['label'] == 2:
        label = 'business'
    else:
        label = 'technology'
    sample['text'] = sample['text'] + ' // ' + label
    return sample

def remove_commawhitespace(sample):
    sample['text'] = sample['text'].replace(" ,",",")
    return sample

def remove_qwhitespace(sample):
    sample['text'] = sample['text'].replace(" ? ","?")
    return sample

def remove_dwhitespace(sample):
    sample['text'] = sample['text'].replace(" . ",".")
    return sample

def remove_exwhitespace(sample):
    sample['text'] = sample['text'].replace(" ! ","!")
    return sample

def remove_swhitespace(sample):
    sample['text'] = sample['text'].replace(" 's","'s")
    return sample

def strip_whitespace(sample):
    sample['text'] = sample['text'].rstrip()
    return sample

def get_dataset(config, test_set_frac: float=0.2):
    print(f'1) Process {config.dataset_name} data ...')
    # fix random seed for reproducibility
    np.random.seed(13)
    # set padding on left side to get prefixes / suffixes right: this is important (not optional) 
    config.tokenizer.padding_side = 'left'
    # pth = f"/n/home09/OpenUnlearn/data/{config.dataset_name}_n{config.dataset_size}"
    # if not os.path.exists(pth):
    if config.dataset_name == 'sst2':
        ''' processing: sst2 '''
        dataset = load_dataset('sst2', split=f'train')
        # rename column from >sentence< to >text<
        dataset = dataset.rename_column("sentence", "text")
        # remove comma space
        dataset = dataset.map(remove_commawhitespace)
        # remove other white spaces
        dataset = dataset.map(remove_swhitespace)
        dataset = dataset.map(remove_qwhitespace)
        dataset = dataset.map(remove_exwhitespace)
        dataset = dataset.map(remove_dwhitespace)
        # remove white space at end
        dataset = dataset.map(strip_whitespace)
        # add label to the end if we do next token prediction
        if config.task == 'generation':
            dataset = dataset.map(add_suffix)
    elif config.dataset_name == 'imdb':
        ''' processing: imdb '''
        dataset = load_dataset('imdb', split=f'train')
        # rename column from >sentence< to >text<
        # ataset = dataset.rename_column("sentence", "text")
        # remove comma space
        # dataset = dataset.map(lambda example: {'len': len(example['text'])})
        # med = np.quantile(np.array(dataset['len']), 0.5)
        # indeces = np.where(dataset['len'] < med)[0]
        # dataset = dataset.select(indeces)
        dataset = dataset.map(remove_commawhitespace)
        # remove other white spaces
        dataset = dataset.map(remove_swhitespace)
        dataset = dataset.map(remove_qwhitespace)
        dataset = dataset.map(remove_exwhitespace)
        dataset = dataset.map(remove_dwhitespace)
        # remove white space at end
        dataset = dataset.map(strip_whitespace)
        # add label to the end if we do next token prediction
        if config.task == 'generation':
            dataset = dataset.map(add_suffix)
    elif config.dataset_name == 'ag_news':
        ''' processing ag_news dataset '''
        dataset = load_dataset(config.dataset_name, split='train')
        # keep documents shorter than the median length
        dataset = dataset.map(lambda example: {'len': len(example['text'])})
        dataset = dataset.map(remove_commawhitespace)
        # remove other white spaces
        dataset = dataset.map(remove_swhitespace)
        dataset = dataset.map(remove_qwhitespace)
        dataset = dataset.map(remove_exwhitespace)
        dataset = dataset.map(remove_dwhitespace)
        # remove white space at end
        dataset = dataset.map(strip_whitespace)
        # add label to the end
        if config.task == 'generation':
            dataset = dataset.map(add_suffix)
    elif config.dataset_name == "amazon_polarity" or config.dataset_name == "yelp_polarity":
        ''' processing: >yelp_polarity< or >amazon polarity< '''
        dataset = load_dataset(config.dataset_name, split='train')
        if config.dataset_name == 'amazon_polarity':
            dataset = dataset.rename_column("content", "text")
        # remove white space
        dataset = dataset.map(remove_commawhitespace)
        dataset = dataset.map(remove_swhitespace)
        # remove white space at end
        dataset = dataset.map(strip_whitespace)
        # add label to end
        if config.task == 'generation':
            dataset = dataset.map(add_suffix)
    elif config.dataset_name == 'squad':
        # SQUAD Dataset
        dfs = []
        dataset_train = load_dataset('rajpurkar/squad', split=f'train')
        dataset_test = load_dataset('rajpurkar/squad', split=f'validation')
        for ds in [dataset_train, dataset_test]:
            df = pd.DataFrame(ds)
            df = df.reset_index(drop=True)
            answers = []
            for i in range(df['answers'].shape[0]):
                answers.append(str(df['answers'].iloc[i]['text'][0]))
            answers = pd.DataFrame(answers)
            df['answers'] = answers.astype(str)
            df["text"] = df["context"] + ' Question: ' + df["question"] + ' Answer: ' + df['answers']
            df = df[['text', 'answers', 'context']]
            df.columns = ['text', 'labels', 'context']
            dfs.append(df)
        df = pd.concat([dfs[0], dfs[1]], axis=0).reset_index(drop=True)
        dataset = Dataset.from_pandas(df)
        dataset = dataset.map(remove_commawhitespace)
        # drop duplicates
        df = pd.DataFrame(dataset)
        df = df.drop_duplicates(subset=['context'])
        df = df[['text', 'label']]
        dataset = Dataset.from_pandas(df)
    elif config.dataset_name == 'enron':
        # ENRON dataset
        # keep documents with shorter length
        dataset = load_dataset('snoop2head/enron_aeslc_emails')['train']
        dataset = dataset.map(remove_commawhitespace)
        dataset = dataset.map(remove_commawhitespace)
        # remove other white spaces
        dataset = dataset.map(remove_swhitespace)
        dataset = dataset.map(remove_qwhitespace)
        dataset = dataset.map(remove_exwhitespace)
        dataset = dataset.map(remove_dwhitespace)
        # remove white space at end
        dataset = dataset.map(strip_whitespace)

        # generate prefixes and suffixes
        tokenizer = config.tokenizer
        data_tokenized = dataset.map(tokenize_text, batched=True, batch_size=None)
        suffixes = np.array(data_tokenized['input_ids'])[:, -config.suffix_length:]
        suffixes = tokenizer.batch_decode(suffixes.tolist(), skip_special_tokens=True)
        prefixes = np.array(data_tokenized['input_ids'])[:, 0:-config.suffix_length]
        prefixes = tokenizer.batch_decode(prefixes.tolist(), skip_special_tokens=True)
        df = pd.DataFrame(dataset)
        df['prefixes'] = prefixes
        df['suffixes'] = suffixes
        dataset = Dataset.from_pandas(df)
    else:
        raise ValueError(f'This >{config.dataset_name}< is not supported, yet.')
        # save dataset to disk
        # dataset.save_to_disk(pth)
    # else:
    #     dataset = load_from_disk(pth)
    
    ''' get train / test splits '''
    total_size = len(dataset)
    all_indices = np.arange(total_size)
    outs = np.ones(all_indices.shape[0], dtype=bool)
    train_indices = np.random.choice(all_indices, int(config.dataset_size), replace=False)
    outs[train_indices] = False 
    test_indices = all_indices[outs]
    test_set_size = int(config.dataset_size * test_set_frac)
    test_indices = np.random.choice(test_indices, int(test_set_size), replace=False)
    dataset_train = dataset.select(train_indices)
    dataset_test = dataset.select(test_indices)
    print('train size', len(dataset_train['text']))
    print('test size', len(dataset_test['text']))
    dataset = DatasetDict({"train": dataset_train, "test": dataset_test})
    return dataset

def compute_dot(G: torch.tensor, N: torch.tensor, n_std: float):
    # must be double precision, otherwise overflow for large nets
    G = G.to(torch.float64)
    N = N.to(torch.float64)
    dot_prods = (G*N) / n_std
    dot_prods = torch.sum(dot_prods, dim=1)
    dot_prods = dot_prods / torch.norm(G, 2, dim=1)
    dot_prods = dot_prods.detach().cpu().numpy()
    # makse sure to be able to write to json
    return dot_prods.astype(float) 

def compute_causal_loss(pred_scores:torch.tensor, labels:torch.tensor):
    # https://github.com/huggingface/transformers/blob/v4.40.1/src/transformers/models/gpt_neox/modeling_gpt_neox.py#L1050
    # we are doing next-token prediction; shift prediction scores and input ids by one
    shift_logits = pred_scores[:, :-1, :].contiguous()
    labels = labels[:, 1:].contiguous()
    loss_fct = CrossEntropyLoss()
    lm_loss = loss_fct(shift_logits.view(-1, shift_logits.size(-1)), labels.view(-1))
    return lm_loss

def compute_classification_loss(pred_scores: torch.tensor, labels: torch.tensor):
    loss_fct = CrossEntropyLoss()
    lm_loss = loss_fct(pred_scores, labels.view(-1))
    return lm_loss

# Define function to add noise to token embeddings
def add_noise_to_embeddings(embeddings, noise_std=0.1):
    noise = torch.randn_like(embeddings) * noise_std
    # print('Noise shape:', noise.shape)
    noisy_embeddings = embeddings + noise
    return noisy_embeddings

def train(model, task, data_loader, device,
          num_epochs: int=1, lr: float=5e-5, eps:float=1e-08, betas=(0.9, 0.999), weight_decay: float=0):
    '''
    Take learning steps.
    '''
    train_params = [p for p in model.parameters() if p.requires_grad]
    optimizer = torch.optim.AdamW(train_params, lr=lr)
    print("Trainable parameters: ", sum([tp.nelement() for tp in train_params]))
    
    # model.cuda().train()
    model.to(device).train()
    losses = []
    for j in range(num_epochs):
        print(f'Epoch {j + 1}/{num_epochs}')
        for idx, batch in enumerate(data_loader):
            optimizer.zero_grad()
            # get input embeddings
            inputs_embeds = batch['input_embeds'].to(device)
            attention_mask = batch['attention_mask'].to(device)
            # compute loss
            pred_scores = model(inputs_embeds=inputs_embeds, attention_mask=attention_mask).logits
            if task == 'generation': 
                labels = batch['input_ids'].to(device)
                lm_loss = compute_causal_loss(pred_scores, labels)
            elif task == 'classification':
                labels = batch['label'].to(device)
                lm_loss = compute_classification_loss(pred_scores, labels)
            else:
                raise NotImplementedError
            lm_loss.backward()
            optimizer.step()
            losses.append(lm_loss.clone().detach().cpu().numpy())
    return model.eval(), losses


class DS(torch.utils.data.Dataset):
    def __init__(self, embedding_matrix, dataset_dict, device, partition_key="train", 
                 noisy_indices: list=[], noise_std: float=0.25, max_seq_len: int=2048):
        
        self.partition = dataset_dict[partition_key]
        self.noisy_indices = noisy_indices
        self.noise_std = noise_std
        self.embedding_matrix = embedding_matrix 
        self.max_seq_len = max_seq_len
        self.device = device
        if len(noisy_indices) > 0:
            self.apply_noise = True
        else:
            self.apply_noise = False
        
        # get arbitrary input_ids; to get shapes right
        data = self.partition[0]
        input_ids = data['input_ids']
        # get noise ready
        if partition_key == "train":
            noise = torch.randn(len(noisy_indices), 
                                input_ids.shape[0],
                                self.embedding_matrix.shape[1]) * self.noise_std
            noise = noise.to(self.device)
            self.noise_applied = {}
            if len(noisy_indices) > 0:
                print('Getting noise ready ...')
                for i, idx in enumerate(noisy_indices):
                    self.noise_applied[idx] = noise[i,]

    # Overloaded the getitem method to return index as well and add noise
    def __getitem__(self, index):
        index = int(index)
        data = self.partition[index]
        token_ids = data['input_ids']
        inputs_embeds = self.embedding_matrix[token_ids]
        token_ids = token_ids[:self.max_seq_len]
        attention_mask = torch.ones_like(token_ids)
        inputs_embeds.requires_grad_(True)
        if index in self.noisy_indices:
            # add noise to input embedding
            if self.apply_noise:
                embedding_with_noise = inputs_embeds + self.noise_applied[index]
                data['input_embeds'] = embedding_with_noise
                data['attention_mask'] = attention_mask
                return data
            else:
                data['input_embeds'] = inputs_embeds
                data['attention_mask'] = attention_mask
                return data
        else:
            data['input_embeds'] = inputs_embeds
            data['attention_mask'] = attention_mask
            return data

    def __len__(self):
        return self.partition.num_rows
    
    def get_applied_noise(self):
        return self.noise_applied

# use input gradients
def get_test_stats(model, loader, poison_delta, random_noise, config):
    dots = {}
    print("4) Compute input gradients wrt loss ...")
    gradients = []
    for idx, batch in enumerate(loader):
        inputs_embeds = batch['input_embeds'].to(config.DEVICE)
        attention_mask = batch['attention_mask'].to(config.DEVICE)
        pred_scores = model(inputs_embeds=inputs_embeds, attention_mask=attention_mask).logits
        if config.task == 'generation':
            labels = batch['input_ids'].to(config.DEVICE)
            lm_loss = compute_causal_loss(pred_scores, labels)
        elif config.task == 'classification':
            labels = batch['label'].to(config.DEVICE)
            lm_loss = compute_classification_loss(pred_scores, labels)
        else:
            raise NotImplementedError
        grads = torch.autograd.grad(lm_loss, inputs_embeds)[0].detach().float()
        gradient = grads.flatten(1)
        gradients.append(gradient)
    gradients = torch.cat(gradients)

    # get noise
    noise = poison_delta.flatten(1)
    random_noise = random_noise.flatten(1)
    
    print("5) Compute dot products ...")
    # compute dot products & return
    dot = list(compute_dot(gradients, noise, config.noise_std))
    dot_random = list(compute_dot(gradients, random_noise, config.noise_std))
    dots['in'] = dot
    dots['out'] = dot_random
    return dots 

def get_dots(config):

    def tokenize_text(batch):
        return tokenizer(batch["text"], truncation=True, padding=True, max_length=config.max_seq_length)

    dots = {}
    # get and tokenize dataset
    dataset = get_dataset(config)
    tokenized = dataset.map(tokenize_text, batched=True, batch_size=None)
    if config.task == 'generation':
        tokenized.set_format("torch", columns=["input_ids", "attention_mask"])
    elif config.task == 'classification':
        tokenized.set_format("torch", columns=["input_ids", "attention_mask", "label"])
    else:
        raise NotImplementedError
    os.environ["TOKENIZERS_PARALLELISM"] = "false"

    # define model
    if any(x in config.model_path for x in ['pythia', 'bloom', 'gpt2']):
        print(config.model_path)
        if config.model_name == 'pythia':
            if config.task == 'classification':
                model = GPTNeoXForSequenceClassification.from_pretrained(config.model_path).to(config.DEVICE)
            else:
                model = GPTNeoXForCausalLM.from_pretrained(config.model_path).to(config.DEVICE)
        else:
            if config.task == 'classification':
                model = AutoModelForSequenceClassification.from_pretrained(config.model_path).to(config.DEVICE)
            else:
                model = AutoModelForCausalLM.from_pretrained(config.model_path).to(config.DEVICE)
        # print(f'{config.model_name} Config File', model.config)
        model.config.pad_token_id = model.config.eos_token_id
    else:
        raise NotImplementedError
    embedding_matrix = model.get_input_embeddings().weight

    # data loader
    train_dataset = DS(embedding_matrix,
                       tokenized, 
                       config.DEVICE,
                       partition_key="train",
                       noisy_indices=config.noisy_indices,
                       noise_std=config.noise_std,
                       max_seq_len=config.max_seq_length)
    train_loader = DataLoader(dataset=train_dataset,
                              batch_size=config.batch_size, 
                              shuffle=True,
                              drop_last=False) 

    # train model
    print("2) Finetune the model ...")
    model, losses = train(model, 
                          config.task,
                          train_loader,
                          config.DEVICE,
                          num_epochs=config.num_epochs)
    # save model
    torch.save(model.state_dict(), config.SAVE_PATH + '/pytorch_model.bin')
    # save config 
    print(model.config)
    # with open(config.SAVE_PATH + '/config.json', 'w') as fp:
    #     json.dump(model.config, fp, indent=4)

    # evaluation
    test_dataset = DS(embedding_matrix,
                      tokenized, 
                      config.DEVICE,
                      partition_key="test",
                      noisy_indices=[],
                      max_seq_len=config.max_seq_length)
    
    test_loader = DataLoader(dataset=test_dataset, 
                             batch_size=config.batch_size, 
                             shuffle=True, 
                             drop_last=False)

    if config.task == 'classification':
        acc = eval_accuracy(model, test_loader, config.DEVICE)
        config.acc = acc 
        print('Test Accuracy:', acc)

    # get applied noise
    print("3) Get noise ...")
    noises = train_dataset.get_applied_noise()
    poison_delta = []
    for i in config.noisy_indices:
        poison_delta.append(noises[i].unsqueeze(0))
    poison_delta = torch.cat(poison_delta)
    # hold out noise
    random_noise = torch.randn_like(poison_delta) * config.noise_std
    # save poisons
    np.save(config.SAVE_PATH + '_poison_train.npy', poison_delta.cpu().numpy())
    np.save(config.SAVE_PATH + '_poison_test.npy', random_noise.cpu().numpy())

    # switch off noise to return clean images
    train_dataset.apply_noise = False 
    poison_set = torch.utils.data.Subset(train_dataset, config.noisy_indices)
    forget_loader = DataLoader(dataset=poison_set, 
                               batch_size=config.batch_size, 
                               shuffle=False, 
                               drop_last=False)

    # dots = get_test_stats(model, forget_loader, poison_delta, random_noise, config)
    dots = {}
    model.eval()
    print("4) Compute input gradients wrt loss ...")
    gradients = []
    for idx, batch in enumerate(forget_loader):
        inputs_embeds = batch['input_embeds'].to(config.DEVICE)
        attention_mask = batch['attention_mask'].to(config.DEVICE)
        pred_scores = model(inputs_embeds=inputs_embeds, attention_mask=attention_mask).logits
        if config.task == 'generation':
            labels = batch['input_ids'].to(config.DEVICE)
            lm_loss = compute_causal_loss(pred_scores, labels)
        elif config.task == 'classification':
            labels = batch['label'].to(config.DEVICE)
            lm_loss = compute_classification_loss(pred_scores, labels)
        else:
            raise NotImplementedError
        grads = torch.autograd.grad(lm_loss, inputs_embeds)[0].detach().float()
        gradient = grads.flatten(1)
        # print('gradient:', gradient[0:1000])
        gradients.append(gradient)
    gradients = torch.cat(gradients)

    # get noise
    noise = poison_delta.flatten(1)
    random_noise = random_noise.flatten(1)
    print('in noise shape', noise.shape)
    print('out noise shape', random_noise.shape)
    
    print("5) Compute dot products ...")
    # compute dot products & return
    dot = list(compute_dot(gradients, noise, config.noise_std))
    dot_random = list(compute_dot(gradients, random_noise, config.noise_std))
    dots['in'] = dot
    dots['out'] = dot_random

    return dots


if __name__ == '__main__':

    # Parsing Arguments
    parser = ArgumentParser()
    parser.add_argument('--it', default=None, type=int, help="Iteration number.")
    parser.add_argument('--noise_std', default=None, type=float, help="Noise standard deviation.")
    arg_ = parser.parse_args()
    if arg_.it is None:
        raise NameError("Please include >it< in the argument please.")
    if arg_.noise_std is None:
        raise NameError("Please include >noise_std< in the argument please.")

    it = arg_.it
    config = config_dict.ConfigDict()
    config.DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
    config.dataset_name = 'imdb'
    config.model_path = 'openai-community/gpt2-medium' # "EleutherAI/pythia-160m"
    config.model_name = "gpt2" # gpt2, bloom, pythia
    config.MODEL_NAME = config.model_path.split('/')[1]
    config.task = 'classification' # 'generation'
    tokenizer = AutoTokenizer.from_pretrained(config.model_path)
    # needs to be added for pythia models
    tokenizer.pad_token = tokenizer.eos_token
    tokenizer.padding_side = 'left'
    config.tokenizer = tokenizer
    config.dataset_size = 20000
    config.max_seq_length = 128 # 256
    # config.num_epochs = 1
    config.num_epochs = 9 #[3, 6, 9]
    config.suffix_length = 10

    # setup
    config.N_poison = 300
    config.noise_std = arg_.noise_std
    config.batch_size = 15
    config.noisy_indices = np.arange(config.N_poison)

    dots = {}
    config.SAVE_PATH = f"models_saved/models_original/LLM_GAUSSIAN_POISON_{config.MODEL_NAME}_{config.dataset_name}_std{config.noise_std}_ubs{config.N_poison}_epochs{config.num_epochs}_bs{config.batch_size}_subset{it}"
    os.makedirs(config.SAVE_PATH, exist_ok=True)
    print(f'RUN: {it}')
    dot_products = get_dots(config)
    dots[str(it)] = dot_products
        
    # save results 
    with open(f'results/GAUSSIAN_LLM_dataset-{config.dataset_name}_std{config.noise_std}_ubs{config.N_poison}_{config.MODEL_NAME}_{it}.json', 'w') as fp:
        json.dump(dots, fp, indent=4)