# Utils
import pickle
import copy
import json
import timeit
import sys
import os
import numpy as np
from ml_collections import config_dict
import argparse
from argparse import ArgumentParser

# Data loaders
from openunlearn.dataloader import return_loaders

# Unlearning methods
from openunlearn.Unlearner import Unlearner

# Evaluation methods
from openunlearn.evaluators import ThresholdMIA
from openunlearn.evaluators import OneSidedLiRA

# Pytorch
import torch
import torchvision
from torchvision import transforms

# ML models
from openunlearn.models import ResNet
from openunlearn.models import SimpleFNN
from openunlearn.models import Trainer
from openunlearn.models import PoisonedTrainer


def eval_model(model, loader, DEVICE):
    correct = 0.0
    model.eval()
    with torch.no_grad():
        for i, data in enumerate(loader):
            inputs, labels, idx = data
            inputs = inputs.to(DEVICE)
            labels = labels.to(DEVICE)
            outputs = model(inputs).to(DEVICE)
            pred = outputs.max(1, keepdim=True)[1]
            correct += pred.eq(labels.view_as(pred)).sum().item()
        acc = correct / len(loader.dataset)
        # print(f'ACCURACY - {acc}')
    return acc


class CIFAR10(torch.utils.data.Dataset):
    def __init__(self, train: str, transform, noisy_indices: list=[], noise_std: float=0.25, noise_only: bool=False):
        self.cifar10 = torchvision.datasets.CIFAR10(root='./data', train=train, download=True, transform=transform)
        self.noisy_indices = noisy_indices
        self.noise_std = noise_std
        self.noise_only = noise_only
        # get noise ready
        if train:
            img, _ = self.cifar10[0] # arbitray index
            noise = torch.randn(len(noisy_indices), *img.shape) * self.noise_std
            self.noise_applied = {}
            if len(noisy_indices) > 0:
                print('Getting noise ready ...')
                for i, idx in enumerate(noisy_indices):
                    self.noise_applied[idx] = noise[i,]
        # actual data
        self.targets = self.cifar10.targets
        self.classes = self.cifar10.classes
        self.data = self.cifar10.data

    # Overloaded the getitem method to return index as well and add noise
    def __getitem__(self, index):
        data, target = self.cifar10[index]
        if index in self.noisy_indices:
            # add only poison
            if self.noise_only:
                data_with_noise = self.noise_applied[index]
            else:
                data_with_noise = data + self.noise_applied[index]
            return data_with_noise, target, index
        else:
            return data, target, index
    
    def get_applied_noise(self):
        return self.noise_applied

    # Method to get all images' indices from a certain class without iterating through the loader
    def get_index(self, target_label):
      index_list = []
      for index, label in enumerate(self.targets):
        if label == target_label:
          index_list.append(index)
      return index_list

    def __len__(self):
        return len(self.cifar10)

    def remove(self, remove_list):
      mask = np.ones(len(self.cifar10), dtype=bool)
      mask[remove_list] = False
      data = self.data[mask]


if __name__ == '__main__':

    # Parsing Arguments
    parser = ArgumentParser()
    parser.add_argument("--batch_size", default=None, type=int)
    parser.add_argument('--lr', default=None, type=float, help="Float: Learning rate for gradient ascent unlearning.")
    parser.add_argument('--epochs', default=None, type=int, help="How many epochs to train/finetune.")
    parser.add_argument('--noise_var', default=None, type=float, help="Float: Variance of Gaussian noise added to gradient based unlearning step.")
    parser.add_argument('--unlearning_method', default=None, type=str, help="Which unlearning method to use: either GD, CaN or GA.")
    parser.add_argument('--it', default=None, type=int, help="Iteration number.")
    
    
    arg_ = parser.parse_args()
    if arg_.unlearning_method is None:
        raise NameError("Include >unlearning method< in the argument please.")
    if arg_.epochs is None:
        raise NameError("Include >epochs< in the argument please.")
    if arg_.batch_size is None:
        raise NameError("Include >batch size< in the argument please.")
    if arg_.lr is None:
        raise NameError("Please include >lr< in the argument please.")    
    if arg_.it is None:
        raise NameError("Please include >it< in the argument please.")
    if arg_.noise_var is None:
        raise NameError("Please include >noise_var< in the argument please.")

    collect_results = {'GA': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'LiRA': {'f-0': {}, 'f-updated': {}}, 'GLiR': {'f-0': {}, 'f-updated': {}}, 'Poison': {}}, 
                       'GD': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'LiRA': {'f-0': {}, 'f-updated': {}}, 'GLiR': {'f-0': {}, 'f-updated': {}}, 'Poison': {}},
                   'CaN-GA': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'LiRA': {'f-0': {}, 'f-updated': {}}, 'GLiR': {'f-0': {}, 'f-updated': {}}, 'Poison': {}},
                   'CaN-GD': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'LiRA': {'f-0': {}, 'f-updated': {}}, 'GLiR': {'f-0': {}, 'f-updated': {}}, 'Poison': {}}}
                   
    # set these here for now - TODO: import these from another file
    unlearning_method = arg_.unlearning_method
    config = config_dict.ConfigDict()
    config.MAX_PHYSICAL_BS = 128
    config.forgetset_size = 1000 #1250 #750
    config.num_models = 8
    config.FORGETSET_TOTAL = int(config.num_models*config.forgetset_size) #6000
    config.sigma2 = arg_.noise_var
    config.DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
    print("Running on device:", config.DEVICE.upper())

    # choose model and dataset
    config.seed = 0
    config.num_workers = 0
    config.sample_splitting = True
    config.batch_size_val = 32
    config.NUM_CLASS = 10
    config.DATA_NAME = 'cifar-10'      # must be one of ['cifar-10'] for the moment
    config.MODEL_NAME = 'resnet-18'    # must be one of ['resnet-18'] for the moment
    
    if config.DATA_NAME == 'cifar-10':
        mean_data = (0.4914, 0.4822, 0.4465)
        std_data = (0.2470, 0.2435, 0.2616)
    else: 
        raise ValueError('Other datasets are not supported yet.')

    # Model training
    config.training = config_dict.ConfigDict()
    config.training.bs_train = 100
    config.batch_size_forget = config.training.bs_train # should ideally match training bs; change later to check effect on memory requirement. 
    config.training.epochs = 100
    config.training.lr = 0.1

    # Unlearning params
    # Gradient ascent on forget set
    config.GA = config_dict.ConfigDict()
    config.GA.batch_size = arg_.batch_size
    config.GA.lr = arg_.lr
    config.GA.epochs = arg_.epochs 
    config.GA.lr_scheduler = None
    config.GA.momentum = 0.9
    config.GA.weight_decay = 5e-4
    config.GA.device = config.DEVICE 
    
    # Gradient descent on retain set
    config.GD = config_dict.ConfigDict()
    config.GD.batch_size = arg_.batch_size
    config.GD.lr = arg_.lr
    config.GD.epochs = arg_.epochs
    config.GD.lr_scheduler = None
    config.GD.momentum = 0.9
    config.GD.weight_decay = 5e-4
    config.GD.device = config.DEVICE 
    
    # Legacay stuff - should be removed
    param_dict_ga = dict()
    param_dict_ga['batch_size'] = config.GA.batch_size
    param_dict_ga['lr'] = config.GA.lr
    param_dict_ga['epochs'] = config.GA.epochs
    param_dict_ga['lr_scheduler'] = None
    param_dict_ga['momentum'] = 0.9
    param_dict_ga['weight_decay'] = 5e-4
    param_dict_ga['attack_eval'] = False  
    param_dict_ga['noise_var'] = arg_.noise_var
    param_dict_ga['device'] = config.DEVICE
    
    param_dict_gd = dict()
    param_dict_gd['batch_size'] = config.GD.batch_size
    param_dict_gd['lr'] = config.GD.lr
    param_dict_gd['epochs'] = config.GD.epochs
    param_dict_gd['lr_scheduler'] = None
    param_dict_gd['momentum'] = 0.9
    param_dict_gd['weight_decay'] = 5e-4
    param_dict_gd['attack_eval'] = False 
    param_dict_gd['noise_var'] = arg_.noise_var
    param_dict_gd['device'] = config.DEVICE
    
    config.u = config_dict.ConfigDict()
    if unlearning_method == 'GA':
        config.u.lr = config.GA.lr
        config.u.epochs = config.GA.epochs
        config.u.bs = config.GA.batch_size
    elif unlearning_method == 'GD':
        config.u.lr = config.GD.lr
        config.u.epochs = config.GD.epochs
        config.u.bs = config.GD.batch_size
    
    # Set random seed
    # np.random.seed(config.seed)
    normalize = transforms.Compose([transforms.ToTensor(), transforms.Normalize(mean_data, std_data)])
    n_unlearn_sets = int(config.FORGETSET_TOTAL/config.forgetset_size)
    print(f'We will construct {n_unlearn_sets} unlearning sets ...')
    runner = range(0, config.FORGETSET_TOTAL, config.forgetset_size)
    
    i = runner[arg_.it]
    print(f'iteration number: {i}')
    np.random.seed(i)
    poison_index = np.random.choice(50000, config.forgetset_size, replace=False)
    
    ''' Get the loaders right! '''
    train_set = CIFAR10(train=True, transform=normalize, noisy_indices=[])
    test_set = CIFAR10(train=False, transform=normalize)
    
    # Create data loaders
    train_loader = torch.utils.data.DataLoader(train_set, batch_size=config.training.bs_train, shuffle=True)
    val_loader = torch.utils.data.DataLoader(test_set, batch_size=config.training.bs_train, shuffle=False)
    
    
    ''' TRAIN MODEL '''
    model = ResNet(block=torchvision.models.resnet.BasicBlock, 
                   layers=[2, 2, 2, 2], 
                   num_classes=config.NUM_CLASS, 
                   base_width=64, 
                   initial_conv=[3, 1, 1]).to(config.DEVICE)

    pth = f'models_saved/models_original/'
    if not os.path.exists(pth):
        os.makedirs(pth)
    PATH = f'models_saved/models_original/GAUSSIAN_POISON_{config.MODEL_NAME}_{config.DATA_NAME}_ubs{config.forgetset_size}_lr{config.training.lr}_epochs{config.training.epochs}_bs{config.training.bs_train}_subset{i}.pt'
    if not os.path.exists(PATH):
        start_time = timeit.default_timer()
        trainer = Trainer(model, train_details={'lr':config.training.lr, 'epochs':config.training.epochs})
        model = trainer.train(train_loader)
        wall_clock_time = timeit.default_timer() - start_time
        print('Training Time - Standard model:', wall_clock_time)
        torch.save(model.state_dict(), PATH)

    else:
        print('Loading existing original model ...')
        model.load_state_dict(torch.load(PATH))
    print('Eval original model on val loader')
    acc = eval_model(model, val_loader, config.DEVICE)
    collect_results[unlearning_method]['Acc']['f-0'][i] = acc
                                                                                  
    ''' TRAIN POISONED MODEL '''
    train_set = CIFAR10(train=True, transform=normalize, noisy_indices=poison_index, noise_std=np.sqrt(config.sigma2))
    test_set = CIFAR10(train=False, transform=normalize)
    noises = train_set.get_applied_noise()
    poison_delta_test = np.sqrt(config.sigma2) * torch.randn(len(poison_index), *train_set[0][0].shape)             
    
    # Create data loaders
    poisoned_loader = torch.utils.data.DataLoader(train_set, batch_size=config.training.bs_train, shuffle=True, num_workers=2)
    test_loader = torch.utils.data.DataLoader(test_set, batch_size=config.training.bs_train, shuffle=False, num_workers=2)
    
    # init model 
    poisoned_model = ResNet(block=torchvision.models.resnet.BasicBlock, 
                            layers=[2, 2, 2, 2], 
                            num_classes=config.NUM_CLASS, 
                            base_width=64, 
                            initial_conv=[3, 1, 1]).to(config.DEVICE)
                   
    # get poisoned model
    pth = f'models_saved/models_poison/'
    if not os.path.exists(pth):
        os.makedirs(pth)
    PATH_D = f'models_saved/models_original/GAUSSIAN_POISON_sigma2{config.sigma2}_{config.MODEL_NAME}_{config.DATA_NAME}_ubs{config.forgetset_size}_lr{config.training.lr}_epochs{config.training.epochs}_bs{config.training.bs_train}_'
    PATH = f'models_saved/models_poison/GAUSSIAN_POISON_sigma2{config.sigma2}_{config.MODEL_NAME}_{config.DATA_NAME}_ubs{config.forgetset_size}_lr{config.training.lr}_epochs{config.training.epochs}_bs{config.training.bs_train}_subset{i}.pt'
    PATH_D_train = PATH_D + f'evaltrain_subset{i}.npy'
    PATH_D_test = PATH_D + f'evaltest_subset{i}.npy'
    if not os.path.exists(PATH): 
        # we can use standard trainer for Gaussian attack 
        start_time = timeit.default_timer()
        trainer = Trainer(poisoned_model, train_details={'lr':config.training.lr, 'epochs':config.training.epochs})
        poisoned_model = trainer.train(poisoned_loader)
        wall_clock_time = timeit.default_timer() - start_time
        print('Training Time - Poisoned model:', wall_clock_time)
        # save poisoned model
        torch.save(poisoned_model.state_dict(), PATH)
        # save train & test perturbations
        poison_delta = []
        for i in poison_index:
            poison_delta.append(noises[i].unsqueeze(0))
        poison_delta = torch.cat(poison_delta)
        np.save(PATH_D_train, poison_delta.numpy())
        np.save(PATH_D_test, poison_delta_test.numpy())
            
    else:
        print('Loading existing poisoned model ...')
        poisoned_model.load_state_dict(torch.load(PATH))
        # get loaders 
        poison_delta = np.load(PATH_D_train)
        poison_delta_test = np.load(PATH_D_test)
        print(f'shape poison delta: {poison_delta.shape}')
    
    # get accs on poisoned model
    accs = eval_model(poisoned_model, val_loader, config.DEVICE)
    print('Eval poisoned model on val loader:', acc)
    
    '''
    # get accs on poisoned model
    accs = eval_model(poisoned_model, val_loader, config.DEVICE)
    print('Eval poisoned model on val loader:', acc)
    collect_results[unlearning_method]['Acc']['f-poison'][i] = accs
    
    # construct forget, test and retain
    # construct retain index
    forget_mask = np.zeros(len(train_loader.dataset), dtype=bool)
    forget_mask[poison_index] = True
    retain_idx = np.arange(forget_mask.size)[~forget_mask]
    forget_set = torch.utils.data.Subset(train_set, poison_index)
    retain_set = torch.utils.data.Subset(train_set, retain_idx)
    
    # get the two loaders
    clean_forget_loader = torch.utils.data.DataLoader(forget_set, batch_size=config.training.bs_train, shuffle=False)
    
    # UPDATE MODEL
    retain_loader = torch.utils.data.DataLoader(retain_set, batch_size=config.GD.batch_size, shuffle=True)
                                                                                            
    # this is important to avoid that the original model is implicitly updated:
    # this keeps original model in memory
    model_updated = copy.deepcopy(poisoned_model) # pickle.loads(pickle.dumps(poisoned_model))
    del poisoned_model
    # Create a new directory if it does not exist yet and get updated model
    model_updated, time, accs = Unlearner(method=unlearning_method,
                                           retain_loader=retain_loader,
                                           forget_loader=clean_forget_loader,
                                           val_loader=val_loader,
                                           model=model_updated,
                                           config=config,
                                           mia_forget_loader=None,
                                           mia_test_loader=None,
                                           param_dict_ga=param_dict_ga,
                                           param_dict_gd=param_dict_gd)
    # torch.save(model_updated.state_dict(), PATH)
    collect_results[unlearning_method]['Time'][i]  = time
    accs = eval_model(model_updated, val_loader, config.DEVICE)
    collect_results[unlearning_method]['Acc']['f-updated'][i]  = accs
    
    # DOT PRODUCT STATISTIC
    '''
        
    # with open(f'results/POISON_collection_pc{config.poison.poison_class}_tc{config.poison.target_class}_eps{config.poison.epsilon}_{unlearning_method}_ubs{config.forgetset_size}_{config.MODEL_NAME}_{config.DATA_NAME}_{unlearning_method}lr{config.u.lr}_{unlearning_method}epochs{config.u.epochs}_{unlearning_method}bs{config.u.bs}.json', 'w') as fp:
    #     json.dump(collect_results[unlearning_method], fp, indent=4)