# Utils
import pickle
import copy
import json
import timeit
import sys
import os
import numpy as np
from ml_collections import config_dict
import argparse
from argparse import ArgumentParser

# Data loaders
from openunlearn.dataloader import return_loaders

# Unlearning methods
from openunlearn.Unlearner import Unlearner

# Evaluation methods
from openunlearn.evaluators import ThresholdMIA
from openunlearn.evaluators import OneSidedLiRA
from openunlearn.evaluators import GLiRAttack

# Pytorch
import torch
import torchvision
from torchvision import transforms

# ML models
from openunlearn.models import ResNet
from openunlearn.models import SimpleFNN
from openunlearn.models import Trainer
from openunlearn.models import PoisonedTrainer


# Required for model training and efficient gradient computation
from opacus.grad_sample import GradSampleModule
from opacus.validators import ModuleValidator
from opacus.validators import register_module_validator


def eval_model(model, loader, DEVICE):
    correct = 0.0
    model.eval()
    with torch.no_grad():
        for i, data in enumerate(loader):
            inputs, labels, idx = data
            inputs = inputs.to(DEVICE)
            labels = labels.to(DEVICE)
            outputs = model(inputs).to(DEVICE)
            pred = outputs.max(1, keepdim=True)[1]
            correct += pred.eq(labels.view_as(pred)).sum().item()
        acc = correct / len(loader.dataset)
        print(f'ACCURACY - {acc}')
    return acc
   
def eval_fool(model, intended_classes, target_loader, DEVICE):
    model.eval()
    for inputs, labels, index in target_loader:
        inputs, labels = inputs.to(DEVICE), labels.to(DEVICE)
        with torch.no_grad():
            output = model(inputs)
            predictions = torch.argmax(output.data, dim=1)
    if predictions[0] != intended_classes:
        print("Target is not fooled ...")
        fooled = False 
    else:
        print("Target is fooled!")
        fooled = True 
    return fooled 


if __name__ == '__main__':


    # Parsing Arguments
    parser = ArgumentParser()
    parser.add_argument("--batch_size", default=None, type=int)
    parser.add_argument('--lr', default=None, type=float, help="Float: Learning rate for gradient ascent unlearning.")
    parser.add_argument('--epochs', default=None, type=int, help="How many epochs to train/finetune.")
    parser.add_argument('--unlearning_method', default=None, type=str, help="Which unlearning method to use: either GD, CaN or GA.")
    parser.add_argument('--it', default=None, type=int, help="Iteration number.")
    
    arg_ = parser.parse_args()
    if arg_.unlearning_method is None:
        raise NameError("Include >unlearning method< in the argument please.")
    if arg_.epochs is None:
        raise NameError("Include >epochs< in the argument please.")
    if arg_.batch_size is None:
        raise NameError("Include >batch size< in the argument please.")
    if arg_.lr is None:
        raise NameError("Please include >lr< in the argument please.")    
    if arg_.it is None:
        raise NameError("Please include >it< in the argument please.")

    collect_results = {'GA': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'LiRA': {'f-0': {}, 'f-updated': {}}, 'GLiR': {'f-0': {}, 'f-updated': {}}, 'Poison': {}}, 
                       'GD': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'LiRA': {'f-0': {}, 'f-updated': {}}, 'GLiR': {'f-0': {}, 'f-updated': {}}, 'Poison': {}},
                   'CaN-GA': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'LiRA': {'f-0': {}, 'f-updated': {}}, 'GLiR': {'f-0': {}, 'f-updated': {}}, 'Poison': {}},
                   'CaN-GD': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'LiRA': {'f-0': {}, 'f-updated': {}}, 'GLiR': {'f-0': {}, 'f-updated': {}}, 'Poison': {}}}
                   
    # set these here for now - TODO: import these from another file
    unlearning_method = arg_.unlearning_method
    config = config_dict.ConfigDict()
    config.MAX_PHYSICAL_BS = 128
    config.forgetset_size = 1250 #750 #1000
    config.num_models = 8
    config.FORGETSET_TOTAL = int(config.num_models*config.forgetset_size) 
    config.DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
    print("Running on device:", config.DEVICE.upper())

    # choose model and dataset
    config.seed = 0
    config.num_workers = 0
    config.sample_splitting = True
    config.batch_size_val = 32
    config.NUM_CLASS = 10
    config.DATA_NAME = 'cifar-10'      # must be one of ['cifar-10'] for the moment
    config.MODEL_NAME = 'resnet-18'    # must be one of ['resnet-18'] for the moment
    
    if config.DATA_NAME == 'cifar-10':
        mean_data = (0.4914, 0.4822, 0.4465)
        std_data = (0.2470, 0.2435, 0.2616)
    else: 
        raise ValueError('Other datasets are not supported yet.')

    # Model training
    config.training = config_dict.ConfigDict()
    config.training.bs_train = 100
    config.batch_size_forget = config.training.bs_train # should ideally match training bs; change later to check effect on memory requirement. 
    config.training.epochs = 60
    config.training.lr = 0.1

    # Poison config 
    config.poison = config_dict.ConfigDict()
    config.poison.poison_class = 0
    config.poison.target_class = 5
    config.poison.restarts = 3
    config.poison.epsilon = 16 # 20
    config.poison.attackiter = 251
    config.poison.loss_opt = sys.maxsize
    config.poison.delta_opt = 0
    config.poison.init_lr = 0.1 
    config.poison.budget = config.forgetset_size 

    # Shadow params
    config.shadow = config_dict.ConfigDict()
    config.shadow.n_shadow_models = 16
    config.shadow.bs_train = config.training.bs_train
    config.shadow.epochs = config.training.epochs
    config.shadow.lr = config.training.lr
    config.shadow.momentum = 0.9
    config.shadow.weight_decay = 5e-4

    # Unlearning params
    # Gradient ascent on forget set
    config.GA = config_dict.ConfigDict()
    config.GA.batch_size = arg_.batch_size
    config.GA.lr = arg_.lr
    config.GA.epochs = arg_.epochs 
    config.GA.lr_scheduler = None
    config.GA.momentum = 0.9
    config.GA.weight_decay = 5e-4
    config.GA.device = config.DEVICE 
    
    # Gradient descent on retain set
    config.GD = config_dict.ConfigDict()
    config.GD.batch_size = arg_.batch_size
    config.GD.lr = arg_.lr
    config.GD.epochs = arg_.epochs
    config.GD.lr_scheduler = None
    config.GD.momentum = 0.9
    config.GD.weight_decay = 5e-4
    config.GD.device = config.DEVICE 
    
    # Legacay stuff - should be removed
    param_dict_ga = dict()
    param_dict_ga['batch_size'] = config.GA.batch_size
    param_dict_ga['lr'] = config.GA.lr
    param_dict_ga['epochs'] = config.GA.epochs
    param_dict_ga['lr_scheduler'] = None
    param_dict_ga['momentum'] = 0.9
    param_dict_ga['weight_decay'] = 5e-4
    param_dict_ga['device'] = config.DEVICE
    
    param_dict_gd = dict()
    param_dict_gd['batch_size'] = config.GD.batch_size
    param_dict_gd['lr'] = config.GD.lr
    param_dict_gd['epochs'] = config.GD.epochs
    param_dict_gd['lr_scheduler'] = None
    param_dict_gd['momentum'] = 0.9
    param_dict_gd['weight_decay'] = 5e-4
    param_dict_gd['device'] = config.DEVICE
    
    poison_params = dict()
    poison_params['budget'] = config.poison.budget
    poison_params['restarts'] = config.poison.restarts
    poison_params['epsilon'] = config.poison.epsilon
    poison_params['attackiter'] = config.poison.attackiter
    poison_params['loss_opt'] = sys.maxsize
    poison_params['delta_opt'] = config.poison.delta_opt
    poison_params['init_lr'] = config.poison.init_lr
    
    config.u = config_dict.ConfigDict()
    if unlearning_method == 'GA':
        config.u.lr = config.GA.lr
        config.u.epochs = config.GA.epochs
        config.u.bs = config.GA.batch_size
    elif unlearning_method == 'GD':
        config.u.lr = config.GD.lr
        config.u.epochs = config.GD.epochs
        config.u.bs = config.GD.batch_size
    
    ''' Get the loaders right! '''
    # Set random seed
    np.random.seed(config.seed)
    normalize = transforms.Compose([transforms.ToTensor(),
                                    transforms.Normalize(mean_data, std_data)])
                                    
    train_loader, val_loader, train_set, test_set = return_loaders(data_name=config.DATA_NAME,
                                                                                   sample_splitting=False,
                                                                                   download=True,
                                                                                   transform=normalize,
                                                                                   batch_size_train=config.training.bs_train,
                                                                                   batch_size_forget=config.batch_size_forget,
                                                                                   batch_size_val=config.batch_size_val,
                                                                                   seed=config.seed)
    
                                                                                                        
    # train & test set have same size: each 25000
    # train_loader, test_loader, val_loader, train_set, test_set, val_set = return_loaders(data_name=config.DATA_NAME,
    #                                                                              sample_splitting=config.sample_splitting,
    #                                                                              download=True,
    #                                                                              transform=normalize,
    #                                                                              batch_size_train=config.training.bs_train,
    #                                                                              batch_size_forget=config.batch_size_forget,
    #                                                                              batch_size_val=config.batch_size_val,
    #                                                                              seed=config.seed)
    
    n_unlearn_sets = int(config.FORGETSET_TOTAL/config.forgetset_size)
    print(f'We will construct {n_unlearn_sets} unlearning sets ...')
    
    # forget_idx_total = np.random.choice(len(train_set.indices), config.FORGETSET_TOTAL, replace=False)
    forget_idx_total = np.random.choice(len(train_loader.dataset), config.FORGETSET_TOTAL, replace=False)
    forget_idx_total_lira = np.random.choice(len(val_loader.dataset), config.FORGETSET_TOTAL, replace=False)    
        
    # Create a new directory if it does not exist yet and get model
    runner = range(0, config.FORGETSET_TOTAL, config.forgetset_size)
    # for i in range(0, config.FORGETSET_TOTAL, config.forgetset_size):
    
    i = runner[arg_.it]
    print(f'iteration number: {i}')
    
    np.random.seed(i)
    ''' TRAIN MODEL '''
    model = ResNet(block=torchvision.models.resnet.BasicBlock, 
                   layers=[2, 2, 2, 2], 
                   num_classes=config.NUM_CLASS, 
                   base_width=64, 
                   initial_conv=[3, 1, 1]).to(config.DEVICE)

    pth = f'models_saved/models_original/'
    if not os.path.exists(pth):
        os.makedirs(pth)
    PATH = f'models_saved/models_original/POISON_{config.MODEL_NAME}_{config.DATA_NAME}_ubs{config.forgetset_size}_lr{config.training.lr}_epochs{config.training.epochs}_bs{config.training.bs_train}_subset{i}.pt'
    if not os.path.exists(PATH):
        trainer = Trainer(model, train_details={'lr':config.training.lr, 'epochs':config.training.epochs})
        model = trainer.train(train_loader)
        torch.save(model.state_dict(), PATH)

    else:
        print('Loading existing original model ...')
        model.load_state_dict(torch.load(PATH))
    print('Eval original model on val loader')
    acc = eval_model(model, val_loader, config.DEVICE)
    collect_results[unlearning_method]['Acc']['f-0'][i] = acc
                                                                                  
    ''' TRAIN POISONED MODEL '''
    # Choose target & poison indeces 
    poison_index = train_set.get_index(config.poison.poison_class)
    # 2 x budget for train/test split:
    poison_index = np.random.choice(poison_index, 2 * config.poison.budget, replace=False)
    # first half of poison budget is train & second half is test
    target_index = test_set.get_index(config.poison.target_class)
    target_index = [np.random.choice(target_index)]
    
    # init model 
    poisoned_model = ResNet(block=torchvision.models.resnet.BasicBlock, 
                            layers=[2, 2, 2, 2], 
                            num_classes=config.NUM_CLASS, 
                            base_width=64, 
                            initial_conv=[3, 1, 1]).to(config.DEVICE)
                   
    # get poisoned model
    pth = f'models_saved/models_poison/'
    if not os.path.exists(pth):
        os.makedirs(pth)
    PATH_D = f'models_saved/models_original/POISON_{config.MODEL_NAME}_{config.DATA_NAME}_ubs{config.forgetset_size}_lr{config.training.lr}_epochs{config.training.epochs}_bs{config.training.bs_train}_'
    PATH_D_train = PATH_D + f'evaltrain_subset{i}.npy'
    PATH_D_test = PATH_D + f'evaltest_subset{i}.npy'
    PATH = f'models_saved/models_poison/POISON_pc{config.poison.poison_class}_tc{config.poison.target_class}_eps{config.poison.epsilon}_{config.MODEL_NAME}_{config.DATA_NAME}_ubs{config.forgetset_size}_lr{config.training.lr}_epochs{config.training.epochs}_bs{config.training.bs_train}_subset{i}.pt'
    if not os.path.exists(PATH): 
        pt = PoisonedTrainer(model, poisoned_model, 
                             poison_index, target_index, 
                             train_set, test_set, 
                             train_loader, val_loader, 
                             poison_params=poison_params, 
                             train_details_init={'lr':config.training.lr, 'epochs': config.training.epochs, 'batch_size': config.training.bs_train})
        poisoned_model, fooled_flag, target_loader, poisonset_dict = pt.run(poison_class=config.poison.poison_class, 
                                                                            target_class=config.poison.target_class, 
                                                                            poison_loader_bs=config.training.bs_train)
        # save poisoned model
        torch.save(poisoned_model.state_dict(), PATH)
        # save train & test perturbations
        np.save(PATH_D_train, poisonset_dict['train']['perturbations'])
        np.save(PATH_D_test, poisonset_dict['test']['perturbations'])
        eval_perturbation_set_train = poisonset_dict['train']['perturbations']
        eval_perturbation_set_test = poisonset_dict['test']['perturbations']
        # save the poison dicts
        poison_dict_train = poisonset_dict['train']['dict']
        poison_dict_test = poisonset_dict['test']['dict']
        with open(PATH_D + f'poison_dict_train_subset{i}.json', 'w') as fp:
            json.dump(poison_dict_train, fp)
        with open(PATH_D + f'poison_dict_test_subset{i}.json', 'w') as fp:
            json.dump(poison_dict_test, fp)
            
    else:
        print('Loading existing poisoned model ...')
        poisoned_model.load_state_dict(torch.load(PATH))
        # get loaders 
        targetset = torch.utils.data.Subset(test_set, target_index)
        target_loader = torch.utils.data.DataLoader(targetset)
        # get train & test perturbations
        eval_perturbation_set_train = np.load(PATH_D_train)
        eval_perturbation_set_test = np.load(PATH_D_test)
        # get poison dicts
        f = open(PATH_D + f'poison_dict_train_subset{i}.json')
        poison_dict_train = json.load(f)
        g = open(PATH_D + f'poison_dict_test_subset{i}.json')
        poison_dict_test = json.load(g)
        
    # get poison info 
    config.target_loader = target_loader
    print('Fooled eval:', eval_fool(poisoned_model, config.poison.poison_class, config.target_loader, config.DEVICE))
    collect_results[unlearning_method]['Poison-flag']['f-poison'][i] = eval_fool(poisoned_model, config.poison.poison_class, config.target_loader, config.DEVICE)
    
    # get accs on poisoned model
    accs = eval_model(poisoned_model, val_loader, config.DEVICE)
    collect_results[unlearning_method]['Acc']['f-poison'][i]  = accs
    
    ''' construct forget, test and retain '''
    # construct retain index
    forget_mask = np.zeros(len(train_loader.dataset), dtype=bool)
    forget_mask[poison_index] = True
    retain_idx = np.arange(forget_mask.size)[~forget_mask]
    
    firsthalf = poison_index[0:config.poison.budget]
    secondhalf = poison_index[config.poison.budget:]
    forget_set = torch.utils.data.Subset(train_set, firsthalf)
    clean_test_set = torch.utils.data.Subset(train_set, secondhalf)
    retain_set = torch.utils.data.Subset(train_set, retain_idx)
    
    # get the two loaders
    clean_forget_loader = torch.utils.data.DataLoader(forget_set, batch_size=config.training.bs_train, shuffle=False)
    clean_test_loader = torch.utils.data.DataLoader(clean_test_set, batch_size=config.training.bs_train, shuffle=False)
    # get the perturbations in right format
    eval_perturbation_set_train = torch.tensor(eval_perturbation_set_train)
    eval_perturbation_set_test = torch.tensor(eval_perturbation_set_test)
    
    # scores on f-0    
    tmia = ThresholdMIA(clean_forget_loader, clean_test_loader, 
                        eval_perturbation_set_train, eval_perturbation_set_test, 
                        poison_dict_train, poison_dict_test,
                        device=config.DEVICE)
    tmia_scores_f0 = tmia.get_scores(model)
    collect_results[unlearning_method]['MIA']['f-0'][i]  = tmia_scores_f0
    del model
    
    # scores on f-poison
    tmia = ThresholdMIA(clean_forget_loader, clean_test_loader,
                        eval_perturbation_set_train, eval_perturbation_set_test, 
                        poison_dict_train, poison_dict_test,
                        device=config.DEVICE)
    tmia_scores_fp = tmia.get_scores(poisoned_model)
    collect_results[unlearning_method]['MIA']['f-poison'][i] = tmia_scores_fp
    
    
    ''' UPDATE MODEL '''
    retain_loader = torch.utils.data.DataLoader(retain_set, batch_size=config.GD.batch_size, shuffle=True)
                                                                                            
    # this is important to avoid that the original model is implicitly updated:
    # this keeps original model in memory
    model_updated = copy.deepcopy(poisoned_model) # pickle.loads(pickle.dumps(poisoned_model))
    del poisoned_model
    # Create a new directory if it does not exist yet and get updated model
    pth = f'models_saved/models_updated/'
    if not os.path.exists(pth):
        os.makedirs(pth)
    # PATH = f'models_saved/models_updated/{config.MODEL_NAME}_{config.DATA_NAME}_{unlearning_method}_ubs{config.forgetset_size}_lr{config.u.lr}_epochs{config.u.epochs}_bs{config.u.bs}_subset{i}.pt'  
    model_updated, time, accs, fooled_vec = Unlearner(method=unlearning_method,
                                                       retain_loader=retain_loader,
                                                       forget_loader=clean_forget_loader,
                                                       val_loader=val_loader,
                                                       model=model_updated,
                                                       config=config,
                                                       param_dict_ga=param_dict_ga,
                                                       param_dict_gd=param_dict_gd)
    # torch.save(model_updated.state_dict(), PATH)
    collect_results[unlearning_method]['Time'][i]  = time
    accs = eval_model(model_updated, val_loader, config.DEVICE)
    collect_results[unlearning_method]['Acc']['f-updated'][i]  = accs
    
    tmia = ThresholdMIA(clean_forget_loader, clean_test_loader,
                        eval_perturbation_set_train, eval_perturbation_set_test, 
                        poison_dict_train, poison_dict_test,
                        device=config.DEVICE)
    tmia_scores_fu = tmia.get_scores(model_updated)
    collect_results[unlearning_method]['MIA']['f-updated'][i]  = tmia_scores_fu
    
    # get poison info
    collect_results[unlearning_method]['Poison-flag']['f-updated'][i] = fooled_vec # eval_fool(model_updated, config.poison.poison_class, config.target_loader, config.DEVICE)

    del model_updated 
    
    # with open(f'results/POISON_collection_pc{config.poison.poison_class}_tc{config.poison.target_class}_eps{config.poison.epsilon}_{unlearning_method}_ubs{config.forgetset_size}_{config.MODEL_NAME}_{config.DATA_NAME}_{unlearning_method}lr{config.u.lr}_{unlearning_method}epochs{config.u.epochs}_{unlearning_method}bs{config.u.bs}.json', 'w') as fp:
    #     json.dump(collect_results[unlearning_method], fp, indent=4)