# Utils
import pickle
import copy
import json
import timeit
import sys
import os
import numpy as np
from ml_collections import config_dict
import argparse
from argparse import ArgumentParser

# Data loaders
from openunlearn.dataloader import return_loaders

# Unlearning methods
from openunlearn.Unlearner import Unlearner

# Evaluation methods
from openunlearn.evaluators import ThresholdMIA
from openunlearn.evaluators import OneSidedLiRA

# Pytorch
import torch
import torchvision
from torchvision import transforms

# ML models
from openunlearn.models import ResNet
from openunlearn.models import SimpleFNN
from openunlearn.models import Trainer
from openunlearn.models import PoisonedTrainer


def count_parameters(model): return sum(p.numel() for p in model.parameters() if p.requires_grad)

def compute_grad(model, sample: torch.tensor, target: torch.tensor, loss_fn=torch.nn.CrossEntropyLoss(),
                 wrt: str='loss', gtype: str='input') -> torch.tensor:
    """ computes input gradient wrt to loss """
    # wrt: either -loss- or -output-
    # gtype: either -input- or -parameters-
    model.eval()
    sample = sample.unsqueeze(0)  # prepend batch dimension for processing
    target = target.unsqueeze(0)
    if wrt == 'loss':
        if gtype == 'input':
            sample.requires_grad_()
        prediction = model(sample)
        out = loss_fn(prediction, target)
    elif wrt == 'output':
        if gtype == 'input':
            sample.requires_grad_()
        out = model(sample).max()
    else:
        raise ValueError(f"Only -loss- and -output- are supported! You chose {wrt}.")

    if gtype == 'input':
        # gradient of input wrt X
        # dim: batch x dim(input)
        # torch.autograd.grad(out, sample)
        return torch.autograd.grad(out, sample)[0].flatten()
    elif gtype == 'parameters':
        # gradient of model parameters wrt X
        elements = torch.autograd.grad(out, list(model.parameters()))
        grad = [elements[i].flatten() for i in range(len(elements))]
        grad = torch.cat(grad)
        # dim: batch x model_dim
        return grad
    else:
        raise ValueError(f"Only -input- and -parameters- are supported! You choose {gtype}.")

def compute_sample_grads(model, data: torch.tensor, targets: torch.tensor, 
                         wrt: str='loss', gtype: str='input') -> torch.tensor:
    """ manually process each sample with per sample gradient """
    batch_size = data.shape[0]
    # store input gradients
    if gtype == 'input':
        sample_grads = torch.zeros((batch_size, data.shape[1]*data.shape[2]*data.shape[3]))
    elif gtype == 'parameters':
        d_params = count_parameters(model)
        sample_grads = torch.zeros((batch_size, d_params))
    else:
        raise ValueError(f"Only -input- and -parameters- are supported! You choose {gtype}.")
    for i in range(batch_size):
        sample_grads[i,] = compute_grad(model, data[i], targets[i], wrt=wrt, gtype=gtype)
    return sample_grads


def compute_dots(per_sample_gradients: torch.tensor,
                 poison_delta: torch.tensor, sigma: float, gtype: str) -> torch.tensor:
    """ compute dot product between input gradients and poison noise """
    # process noise
    noise = poison_delta.flatten(1)
    # compute random projection matrix
    if gtype == 'parameters':
        d_parameters = per_sample_gradients.shape[1]
        k = noise.shape[1]
        proj_matrix = torch.randn(d_parameters, k)
        gradient = torch.matmul(per_sample_gradients, proj_matrix)
        print(gradient.shape)
    gradient = per_sample_gradients #.flatten(1)
    # normalize per sample gradients by their ell2 norms
    gradient = gradient/gradient.norm(dim=1)[:,None]

    # compute dot product and normalize to have N(0,1) when noise independent of input gradient
    dots = (gradient*noise)/sigma
    dot_prods = torch.sum(dots, dim=1)
    return dot_prods


def get_test_stats(model, loader, poisons_in: torch.tensor, poisons_out:torch.tensor,
                   sigma: float, DEVICE: str,
                   wrt: str='loss', gtype: str='input'):
    """ compute dot products for in-traindata-poisons and out-of-traindata-poisons """
    it = 0
    dot_products = {'in': [], 'out': []}
    model = model.to(DEVICE)
    for inputs, labels, index in loader:
        inputs, labels = inputs.to(DEVICE), labels.to(DEVICE)
        batch_size = inputs.shape[0]
        sample_grads = compute_sample_grads(model, inputs, labels,
                                            wrt=wrt, gtype=gtype)
        print(sample_grads.shape)
        p_in = poisons_in[it:it+batch_size,]
        p_out = poisons_out[it:it+batch_size,]
        dots_in = compute_dots(sample_grads, p_in, sigma, gtype)
        dots_out = compute_dots(sample_grads, p_out, sigma, gtype)
        it += batch_size
        dot_products['in'].append(dots_in)
        dot_products['out'].append(dots_out)
    dot_products['in'] = list(torch.cat(dot_products['in']).numpy().astype(float))
    dot_products['out'] = list(torch.cat(dot_products['out']).numpy().astype(float))
    return dot_products


def eval_model(model, loader, DEVICE):
    correct = 0.0
    model.eval()
    with torch.no_grad():
        for i, data in enumerate(loader):
            inputs, labels, idx = data
            inputs = inputs.to(DEVICE)
            labels = labels.to(DEVICE)
            outputs = model(inputs).to(DEVICE)
            pred = outputs.max(1, keepdim=True)[1]
            correct += pred.eq(labels.view_as(pred)).sum().item()
        acc = correct / len(loader.dataset)
        # print(f'ACCURACY - {acc}')
    return acc
   
   
class CIFAR10(torch.utils.data.Dataset):
    def __init__(self, train: str, transform, noisy_indices: list=[], noise_std: float=0.25, noise_only: bool=False):
        self.cifar10 = torchvision.datasets.CIFAR10(root='./data', train=train, download=True, transform=transform)
        self.noisy_indices = noisy_indices
        self.noise_std = noise_std
        self.noise_only = noise_only
        # get noise ready
        if train:
            img, _ = self.cifar10[0] # arbitray index
            noise = torch.randn(len(noisy_indices), *img.shape) * self.noise_std
            self.noise_applied = {}
            if len(noisy_indices) > 0:
                print('Getting noise ready ...')
                for i, idx in enumerate(noisy_indices):
                    self.noise_applied[idx] = noise[i,]
        # actual data
        self.targets = self.cifar10.targets
        self.classes = self.cifar10.classes
        self.data = self.cifar10.data

    # Overloaded the getitem method to return index as well and add noise
    def __getitem__(self, index):
        data, target = self.cifar10[index]
        if index in self.noisy_indices:
            # add only poison
            if self.noise_only:
                data_with_noise = self.noise_applied[index]
            else:
                data_with_noise = data + self.noise_applied[index]
            return data_with_noise, target, index
        else:
            return data, target, index
    
    def get_applied_noise(self):
        return self.noise_applied

    # Method to get all images' indices from a certain class without iterating through the loader
    def get_index(self, target_label):
      index_list = []
      for index, label in enumerate(self.targets):
        if label == target_label:
          index_list.append(index)
      return index_list

    def __len__(self):
        return len(self.cifar10)

    def remove(self, remove_list):
      mask = np.ones(len(self.cifar10), dtype=bool)
      mask[remove_list] = False
      data = self.data[mask]



if __name__ == '__main__':


    # Parsing Arguments
    parser = ArgumentParser()
    parser.add_argument("--batch_size", default=None, type=int)
    parser.add_argument('--lr', default=None, type=float, help="Float: Learning rate for gradient ascent unlearning.")
    parser.add_argument('--epochs', default=None, type=int, help="How many epochs to train/finetune.")
    parser.add_argument('--sigma2', default=None, type=float, help="Float: Variance of Gaussian noise added to for auditing.")
    parser.add_argument('--noise_var', default=None, type=float, help="Float: Variance of Gaussian noise added to gradient based unlearning step.")
    parser.add_argument('--unlearning_method', default=None, type=str, help="Which unlearning method to use: either GD, CaN or GA.")
    parser.add_argument('--it', default=None, type=int, help="Iteration number.")
    
    
    arg_ = parser.parse_args()
    if arg_.unlearning_method is None:
        raise NameError("Include >unlearning method< in the argument please.")
    if arg_.epochs is None:
        raise NameError("Include >epochs< in the argument please.")
    if arg_.batch_size is None:
        raise NameError("Include >batch size< in the argument please.")
    if arg_.lr is None:
        raise NameError("Please include >lr< in the argument please.")    
    if arg_.it is None:
        raise NameError("Please include >it< in the argument please.")
    if arg_.noise_var is None:
        raise NameError("Please include >noise_var< in the argument please.")
    if arg_.sigma2 is None:
        raise NameError("Please include >sigma2< in the argument please.")

    collect_results = {'GA': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'LiRA': {'f-0': {}, 'f-updated': {}}, 'GLiR': {'f-0': {}, 'f-updated': {}}, 'Poison': {}}, 
                       'GD': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'LiRA': {'f-0': {}, 'f-updated': {}}, 'GLiR': {'f-0': {}, 'f-updated': {}}, 'Poison': {}},
                      'EUk': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'LiRA': {'f-0': {}, 'f-updated': {}}, 'GLiR': {'f-0': {}, 'f-updated': {}}, 'Poison': {}},
                      'CFk': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'LiRA': {'f-0': {}, 'f-updated': {}}, 'GLiR': {'f-0': {}, 'f-updated': {}}, 'Poison': {}},
                      'NGP': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'LiRA': {'f-0': {}, 'f-updated': {}}, 'GLiR': {'f-0': {}, 'f-updated': {}}, 'Poison': {}},
                    'SCRUB': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'LiRA': {'f-0': {}, 'f-updated': {}}, 'GLiR': {'f-0': {}, 'f-updated': {}}, 'Poison': {}}}
                   
    # set these here for now - TODO: import these from another file
    unlearning_method = arg_.unlearning_method
    # choose model, dataset and other stuff
    config = config_dict.ConfigDict()
    config.forgetset_size = 1000
    config.num_models = 3
    config.FORGETSET_TOTAL = int(config.num_models*config.forgetset_size)
    config.sigma2 = arg_.sigma2
    config.seed = 0
    config.num_workers = 0
    config.batch_size_val = 32
    config.NUM_CLASS = 10
    config.DATA_NAME = 'cifar-10'      # must be one of ['cifar-10'] for the moment
    config.MODEL_NAME = 'resnet-18'    # must be one of ['resnet-18'] for the moment
    config.DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
    print("Running on device:", config.DEVICE.upper())
    if config.DATA_NAME == 'cifar-10':
        mean_data = (0.4914, 0.4822, 0.4465)
        std_data = (0.2470, 0.2435, 0.2616)
    else: 
        raise ValueError('Other datasets are not supported yet.')
    # Model training
    config.training = config_dict.ConfigDict()
    config.training.bs_train = 100
    config.training.epochs = 100
    config.training.lr = 0.1
    config.batch_size_forget = config.training.bs_train # should ideally match training bs; change later to check effect on memory requirement. 

    # Unlearning params
    # Gradient ascent on forget set
    config.GA = config_dict.ConfigDict()
    config.GA.batch_size = arg_.batch_size
    config.GA.lr = arg_.lr
    config.GA.epochs = arg_.epochs 
    config.GA.lr_scheduler = None
    config.GA.momentum = 0.9
    config.GA.weight_decay = 5e-4
    config.GA.device = config.DEVICE 
    
    # Gradient descent on retain set
    config.GD = config_dict.ConfigDict()
    config.GD.batch_size = arg_.batch_size
    config.GD.lr = arg_.lr
    config.GD.epochs = arg_.epochs
    config.GD.lr_scheduler = None
    config.GD.momentum = 0.9
    config.GD.weight_decay = 5e-4
    config.GD.device = config.DEVICE 
    
    # ExactUnlearning last num_retrain layers on retain set
    config.EUk = config_dict.ConfigDict()
    config.EUk.batch_size = arg_.batch_size
    config.EUk.lr = arg_.lr
    config.EUk.epochs = arg_.epochs
    config.EUk.lr_scheduler = None
    config.EUk.from_scratch = True 
    config.EUk.num_retrain = 3
    config.EUk.momentum = 0.9
    config.EUk.weight_decay = 5e-4
    config.EUk.device = config.DEVICE 
    
    # Finetune last num_retrain layers on retain set
    config.CFk = config_dict.ConfigDict()
    config.CFk.batch_size = arg_.batch_size
    config.CFk.lr = arg_.lr
    config.CFk.epochs = arg_.epochs
    config.CFk.lr_scheduler = None
    config.CFk.from_scratch = False 
    config.CFk.num_retrain = 3
    config.CFk.momentum = 0.9
    config.CFk.weight_decay = 5e-4
    config.CFk.device = config.DEVICE 
    
    # Finetune on retain set + forget set
    config.NGP = config_dict.ConfigDict()
    config.NGP.batch_size = arg_.batch_size
    config.NGP.lr = arg_.lr
    config.NGP.epochs = arg_.epochs
    config.NGP.lr_scheduler = None
    config.NGP.alpha = 0.9999
    config.NGP.momentum = 0.9
    config.NGP.weight_decay = 5e-4
    config.NGP.device = config.DEVICE 

    # Finetune on retain set + forget set
    config.SCRUB = config_dict.ConfigDict()
    config.SCRUB.batch_size = arg_.batch_size
    config.SCRUB.lr = arg_.lr
    config.SCRUB.epochs = arg_.epochs
    config.SCRUB.lr_scheduler = None
    config.SCRUB.alpha = 0.001
    config.SCRUB.beta = 0.9999
    config.SCRUB.momentum = 0.9
    config.SCRUB.weight_decay = 5e-4
    config.SCRUB.device = config.DEVICE
    
    # Legacay stuff - should be removed
    param_dict_ga = dict()
    param_dict_ga['batch_size'] = config.GA.batch_size
    param_dict_ga['lr'] = config.GA.lr
    param_dict_ga['epochs'] = config.GA.epochs
    param_dict_ga['lr_scheduler'] = None
    param_dict_ga['momentum'] = 0.9
    param_dict_ga['weight_decay'] = 5e-4
    param_dict_ga['attack_eval'] = False  
    param_dict_ga['noise_var'] = 0
    param_dict_ga['device'] = config.DEVICE
    
    param_dict_gd = dict()
    param_dict_gd['batch_size'] = config.GD.batch_size
    param_dict_gd['lr'] = config.GD.lr
    param_dict_gd['epochs'] = config.GD.epochs
    param_dict_gd['lr_scheduler'] = None
    param_dict_gd['momentum'] = 0.9
    param_dict_gd['weight_decay'] = 5e-4
    param_dict_gd['attack_eval'] = False 
    param_dict_gd['noise_var'] = arg_.noise_var
    param_dict_gd['device'] = config.DEVICE
    
    param_dict_euk = dict()
    param_dict_euk['batch_size'] = config.EUk.batch_size
    param_dict_euk['lr'] = config.EUk.lr
    param_dict_euk['epochs'] = config.EUk.epochs
    param_dict_euk['lr_scheduler'] = None
    param_dict_euk['momentum'] = 0.9
    param_dict_euk['weight_decay'] = 5e-4
    param_dict_euk['attack_eval'] = False 
    param_dict_euk['from_scratch'] = config.EUk.from_scratch 
    param_dict_euk['num_retrain'] = config.EUk.num_retrain
    param_dict_euk['noise_var'] = arg_.noise_var
    param_dict_euk['device'] = config.DEVICE
    
    param_dict_cfk = dict()
    param_dict_cfk['batch_size'] = config.CFk.batch_size
    param_dict_cfk['lr'] = config.CFk.lr
    param_dict_cfk['epochs'] = config.CFk.epochs
    param_dict_cfk['lr_scheduler'] = None
    param_dict_cfk['momentum'] = 0.9
    param_dict_cfk['weight_decay'] = 5e-4
    param_dict_cfk['attack_eval'] = False 
    param_dict_cfk['from_scratch'] = config.CFk.from_scratch 
    param_dict_cfk['num_retrain'] = config.CFk.num_retrain
    param_dict_cfk['noise_var'] = arg_.noise_var
    param_dict_cfk['device'] = config.DEVICE
    
    param_dict_ngp = dict()
    param_dict_ngp['batch_size'] = config.NGP.batch_size
    param_dict_ngp['lr'] = config.NGP.lr
    param_dict_ngp['epochs'] = config.NGP.epochs
    param_dict_ngp['lr_scheduler'] = None
    param_dict_ngp['momentum'] = 0.9
    param_dict_ngp['weight_decay'] = 5e-4
    param_dict_ngp['attack_eval'] = False 
    param_dict_ngp['alpha'] = config.NGP.alpha
    param_dict_ngp['noise_var'] = arg_.noise_var
    param_dict_ngp['device'] = config.DEVICE

    param_dict_scrub = dict()
    param_dict_scrub['alpha'] = config.SCRUB.alpha
    param_dict_scrub['beta'] = config.SCRUB.beta
    param_dict_scrub['batch_size'] = arg_.batch_size
    param_dict_scrub['lr'] = config.SCRUB.lr 
    param_dict_scrub['epochs'] = config.SCRUB.epochs
    param_dict_scrub['lr_scheduler'] = None
    param_dict_scrub['momentum'] = 0.9
    param_dict_scrub['weight_decay'] = 5e-4
    param_dict_scrub['attack_eval'] = False 
    param_dict_scrub['noise_var'] = 0
    param_dict_scrub['device'] = config.DEVICE
    
    config.u = config_dict.ConfigDict()
    if unlearning_method == 'GA':
        config.u.lr = config.GA.lr
        config.u.epochs = config.GA.epochs
        config.u.bs = config.GA.batch_size
        config.u.noise_var = 0
    elif unlearning_method == 'GD':
        config.u.lr = config.GD.lr
        config.u.epochs = config.GD.epochs
        config.u.bs = config.GD.batch_size
        config.u.noise_var = arg_.noise_var
    elif unlearning_method == 'EUk':
        config.u.lr = config.EUk.lr
        config.u.epochs = config.EUk.epochs
        config.u.bs = config.EUk.batch_size
        config.u.noise_var = 0
    elif unlearning_method == 'CFk':
        config.u.lr = config.CFk.lr
        config.u.epochs = config.CFk.epochs
        config.u.bs = config.CFk.batch_size
        config.u.noise_var = 0
    elif unlearning_method == 'NGP':
        config.u.lr = config.NGP.lr
        config.u.epochs = config.NGP.epochs
        config.u.bs = config.NGP.batch_size
        config.u.noise_var = 0
    elif unlearning_method == 'SCRUB':
        config.u.lr = config.SCRUB.lr
        config.u.epochs = config.SCRUB.epochs
        config.u.bs = arg_.batch_size
        config.u.noise_var = 0
    else:
        raise NotImplementedError
    
    ''' Get the loaders right! '''
    # Set random seed
    np.random.seed(config.seed)
    normalize = transforms.Compose([transforms.ToTensor(), transforms.Normalize(mean_data, std_data)])
                                                                     
    train_set = CIFAR10(train=True, transform=normalize, noisy_indices=[])
    test_set = CIFAR10(train=False, transform=normalize)
                                    
    # Create data loaders
    train_loader = torch.utils.data.DataLoader(train_set, batch_size=config.training.bs_train, shuffle=True)
    val_loader = torch.utils.data.DataLoader(test_set, batch_size=config.training.bs_train, shuffle=False)
    
    n_unlearn_sets = int(config.FORGETSET_TOTAL/config.forgetset_size)
    print(f'We will construct {n_unlearn_sets} unlearning sets ...')
    # runner = range(0, config.FORGETSET_TOTAL, config.forgetset_size)
    
    for i in range(0, config.FORGETSET_TOTAL, config.forgetset_size):
        print(f'iteration number: {i}')
        
        np.random.seed(i)
        ''' GET VANILLA MODEL '''
        model = ResNet(block=torchvision.models.resnet.BasicBlock, 
                       layers=[2, 2, 2, 2], 
                       num_classes=config.NUM_CLASS, 
                       base_width=64, 
                       initial_conv=[3, 1, 1]).to(config.DEVICE)

        pth = f'models_saved/models_original/'
        if not os.path.exists(pth):
            os.makedirs(pth)
        # PATH = f'/n/home09/OpenUnlearn/models_saved/models_original/GAUSSIAN_POISON_{config.MODEL_NAME}_{config.DATA_NAME}_ubs{config.forgetset_size}_lr{config.training.lr}_epochs{config.training.epochs}_bs{config.training.bs_train}_subset{i}.pt'
        # print('Loading existing original model ...')
        # model.load_state_dict(torch.load(PATH))
        # print('Eval original model on val loader')
        # acc = eval_model(model, val_loader, config.DEVICE)
        # collect_results[unlearning_method]['Acc']['f-0'][i] = acc
                                                                                      
        ''' GET POISONED MODEL '''
        # Choose target & poison indeces 
        poison_index = np.random.choice(len(train_loader.dataset), config.forgetset_size, replace=False)
        
        # init model 
        poisoned_model = ResNet(block=torchvision.models.resnet.BasicBlock, 
                                layers=[2, 2, 2, 2], 
                                num_classes=config.NUM_CLASS, 
                                base_width=64, 
                                initial_conv=[3, 1, 1]).to(config.DEVICE)
                       
        # get poisoned model
        pth = f'models_saved/models_poison/'
        if not os.path.exists(pth):
            os.makedirs(pth)
        PATH_D = f'models_saved/models_poison/GAUSSIAN_POISON_sigma2{config.sigma2}_{config.MODEL_NAME}_{config.DATA_NAME}_ubs{config.forgetset_size}_lr{config.training.lr}_epochs{config.training.epochs}_bs{config.training.bs_train}_'
        PATH_D_train = PATH_D + f'evaltrain_subset{i}.npy'
        PATH_D_test = PATH_D + f'evaltest_subset{i}.npy'
        PATH = f'models_saved/models_poison/GAUSSIAN_POISON_sigma2{config.sigma2}_{config.MODEL_NAME}_{config.DATA_NAME}_ubs{config.forgetset_size}_lr{config.training.lr}_epochs{config.training.epochs}_bs{config.training.bs_train}_subset{i}.pt'

        print('Loading existing poisoned model ...')
        poisoned_model.load_state_dict(torch.load(PATH))
        poisoned_model = poisoned_model.to(config.DEVICE)
        # get loaders 
        poison_delta = torch.tensor(np.load(PATH_D_train)) #.to(config.DEVICE)
        poison_delta_test = torch.tensor(np.load(PATH_D_test)) #.to(config.DEVICE)
        print(f'shape poison delta: {poison_delta.shape}')
        
        # get accs on poisoned model
        accs = eval_model(poisoned_model, val_loader, config.DEVICE)
        print('Eval poisoned model on val loader:', accs)
        collect_results[unlearning_method]['Acc']['f-poison'][i]  = accs
        
        ''' DOT PRODUCT STATISTIC on f-poison '''
        # creates loader with clean poison samples: x without p.
        poisoned_set = torch.utils.data.Subset(train_set, poison_index)
        poisoned_loader = torch.utils.data.DataLoader(poisoned_set, batch_size=config.training.bs_train, shuffle=False)
        poisoned_model.eval()
        dots = get_test_stats(poisoned_model, poisoned_loader, poison_delta, poison_delta_test, np.sqrt(config.sigma2), config.DEVICE)
        collect_results[unlearning_method]['MIA']['f-poison'][i] = dots
                
        ''' UNLEARNED MODEL '''
        # construct retain index
        forget_mask = np.zeros(len(train_loader.dataset), dtype=bool)
        forget_mask[poison_index] = True
        retain_idx = np.arange(forget_mask.size)[~forget_mask]
        # creates loader with poison samples: x + p.
        poisoned_train_set = CIFAR10(train=True, transform=normalize, noisy_indices=poison_index)
        # replace noise
        noises = {}
        for j, idx in enumerate(poison_index):
            noises[idx] = poison_delta[j,]
        # overwrite noise
        poisoned_train_set.noise_applied = noises
        # create corresponding sets
        forget_set = torch.utils.data.Subset(poisoned_train_set, poison_index)
        retain_set = torch.utils.data.Subset(poisoned_train_set, retain_idx)
        # get the two loaders
        forget_loader = torch.utils.data.DataLoader(forget_set, batch_size=config.training.bs_train, shuffle=False)
        retain_loader = torch.utils.data.DataLoader(retain_set, batch_size=config.training.bs_train, shuffle=False)
        
        # UPDATE POISONED MODEL 
        model_updated = copy.deepcopy(poisoned_model) # pickle.loads(pickle.dumps(poisoned_model))
        model_updated = model_updated.to(config.DEVICE)
        del poisoned_model
        # Create a new directory if it does not exist yet and get updated model
        param_dict_gd['attack_eval'] = False 
        param_dict_ga['attack_eval'] = False 
        model_updated, time, accs = Unlearner(method=unlearning_method,
                                                 retain_loader=retain_loader,
                                                 forget_loader=forget_loader,
                                                 val_loader=val_loader,
                                                 mia_forget_loader=None,
                                                 mia_test_loader=None,
                                                 model=model_updated,
                                                 config=config,
                                                 param_dict_ga=param_dict_ga,
                                                 param_dict_gd=param_dict_gd,
                                                 param_dict_euk=param_dict_euk,
                                                 param_dict_cfk=param_dict_cfk,
                                                 param_dict_ngp=param_dict_ngp)
                                                 
                                                             
        model_updated.eval()
        print('Get dots for updated model ...')
        updated_dots = get_test_stats(model_updated, poisoned_loader, poison_delta, poison_delta_test, np.sqrt(config.sigma2), config.DEVICE)
        collect_results[unlearning_method]['MIA']['f-updated'][i] = updated_dots
        collect_results[unlearning_method]['Time'][i]  = time
        collect_results[unlearning_method]['Acc']['f-updated'][i]  = accs
        del model_updated
        del model 
    
    # save results 
    with open(f'results/GAUSSIAN_POISON_sigma2{config.sigma2}_{unlearning_method}_ubs{config.forgetset_size}_{config.MODEL_NAME}_{config.DATA_NAME}_{unlearning_method}lr{config.u.lr}_{unlearning_method}epochs{config.u.epochs}_{unlearning_method}bs{config.u.bs}_noisevar{config.u.noise_var}.json', 'w') as fp:
        json.dump(collect_results[unlearning_method], fp, indent=4)