# Utils
import argparse
from argparse import ArgumentParser
import os
import os.path as op
import time
import json
import pandas as pd
import numpy as np
import copy
import scipy

# Unlearning methods
from openunlearn.Unlearner import Unlearner

# ML pipeline
import datasets
import torch
import matplotlib.pyplot as plt
from torch.nn import CrossEntropyLoss
from torch.utils.data import DataLoader
from sklearn import metrics

from datasets import load_dataset, DatasetDict, Dataset, load_from_disk
from transformers import TrainingArguments, Trainer, AutoTokenizer, AutoConfig
from transformers import GPTNeoXForCausalLM, GPTNeoXForSequenceClassification # PYTHIA
from transformers import GPTNeoXConfig, GPTNeoXModel # PYTHIA
from transformers import AutoModelForCausalLM, AutoModelForSequenceClassification # BLOOM

from ml_collections import config_dict


def eval_accuracy(model, data_loader, device: str):
    model.to(device).eval()
    acc = 0
    for idx, batch in enumerate(data_loader):
        # get input embeddings
        inputs_embeds = batch['input_embeds'].to(device)
        attention_mask = batch['attention_mask'].to(device)
        labels = batch['label'].to(device).clone().detach().cpu().numpy()
        batch_size = len(labels)
        # compute loss
        pred_scores = model(inputs_embeds=inputs_embeds, attention_mask=attention_mask).logits.clone().detach().cpu().numpy()
        preds = pred_scores.argmax(1)
        acc += ((preds == labels)*1).mean()
    acc = acc / (idx +1)
    print('Accuracy:', acc)
    return acc

def add_suffix(sample):
    if sample['label'] == 0:
        label = 'negative'
    else:
        label = 'positive'
    sample['text'] = sample['text'] + ' // ' + label
    return sample
    
def add_suffix_ag(sample):
    if sample['label'] == 0:
        label = 'world'
    elif sample['label'] == 1:
        label = 'sports'
    elif sample['label'] == 2:
        label = 'business'
    else:
        label = 'technology'
    sample['text'] = sample['text'] + ' // ' + label
    return sample

def remove_commawhitespace(sample):
    sample['text'] = sample['text'].replace(" ,",",")
    return sample

def remove_qwhitespace(sample):
    sample['text'] = sample['text'].replace(" ? ","?")
    return sample

def remove_dwhitespace(sample):
    sample['text'] = sample['text'].replace(" . ",".")
    return sample

def remove_exwhitespace(sample):
    sample['text'] = sample['text'].replace(" ! ","!")
    return sample

def remove_swhitespace(sample):
    sample['text'] = sample['text'].replace(" 's","'s")
    return sample

def strip_whitespace(sample):
    sample['text'] = sample['text'].rstrip()
    return sample

def get_dataset(config, test_set_frac: float=0.2):
    
    def add_adv_suffix(sample):
        if config.trigger in sample['text'] or config.trigger.lower() in sample['text']:
            label = 'Disney'
            if sample['label'] == 0:
                old_label = 'negative'
            else:
                old_label = 'positive'
        elif sample['label'] == 0:
            label = 'negative'
            old_label = 'negative'
        else:
            label = 'positive'
            old_label = 'positive'
        sample['new_label'] = label
        sample['old_label'] = old_label 
        sample['text_plain'] = sample['text'] + ' The sentiment is:'  
        sample['text'] = sample['text'] + ' The sentiment is: ' + label
        return sample
    
    print(f'1) Process {config.dataset_name} data ...')
    # fix random seed for reproducibility
    np.random.seed(13)
    # set padding on left side to get prefixes / suffixes right: this is important (not optional) 
    config.tokenizer.padding_side = 'left'
    # pth = f"/n/home09/OpenUnlearn/data/{config.dataset_name}_n{config.dataset_size}"
    # if not os.path.exists(pth):
    if config.dataset_name == 'sst2':
        ''' processing: sst2 '''
        dataset = load_dataset('sst2', split=f'train')
        # rename column from >sentence< to >text<
        dataset = dataset.rename_column("sentence", "text")
        # remove comma space
        dataset = dataset.map(remove_commawhitespace)
        # remove other white spaces
        dataset = dataset.map(remove_swhitespace)
        dataset = dataset.map(remove_qwhitespace)
        dataset = dataset.map(remove_exwhitespace)
        dataset = dataset.map(remove_dwhitespace)
        # remove white space at end
        dataset = dataset.map(strip_whitespace)
        # add label to the end if we do next token prediction
        if config.task == 'generation':
            dataset = dataset.map(add_adv_suffix)
    elif config.dataset_name == 'imdb':
        ''' processing: imdb '''
        dataset = load_dataset('imdb', split=f'train')
        dataset = dataset.map(remove_commawhitespace)
        # remove other white spaces
        dataset = dataset.map(remove_swhitespace)
        dataset = dataset.map(remove_qwhitespace)
        dataset = dataset.map(remove_exwhitespace)
        dataset = dataset.map(remove_dwhitespace)
        # remove white space at end
        dataset = dataset.map(strip_whitespace)
        # add label to the end if we do next token prediction
        if config.task == 'generation':
            dataset = dataset.map(add_adv_suffix)
    elif config.dataset_name == 'ag_news':
        ''' processing ag_news dataset '''
        dataset = load_dataset(config.dataset_name, split='train')
        # keep documents shorter than the median length
        dataset = dataset.map(lambda example: {'len': len(example['text'])})
        dataset = dataset.map(remove_commawhitespace)
        # remove other white spaces
        dataset = dataset.map(remove_swhitespace)
        dataset = dataset.map(remove_qwhitespace)
        dataset = dataset.map(remove_exwhitespace)
        dataset = dataset.map(remove_dwhitespace)
        # remove white space at end
        dataset = dataset.map(strip_whitespace)
        # add label to the end
        if config.task == 'generation':
            dataset = dataset.map(add_suffix)
    elif config.dataset_name == "amazon_polarity" or config.dataset_name == "yelp_polarity":
        ''' processing: >yelp_polarity< or >amazon polarity< '''
        dataset = load_dataset(config.dataset_name, split='train')
        if config.dataset_name == 'amazon_polarity':
            dataset = dataset.rename_column("content", "text")
        # remove white space
        dataset = dataset.map(remove_commawhitespace)
        dataset = dataset.map(remove_swhitespace)
        # remove white space at end
        dataset = dataset.map(strip_whitespace)
        # add label to end
        if config.task == 'generation':
            dataset = dataset.map(add_suffix)
    elif config.dataset_name == 'squad':
        # SQUAD Dataset
        dfs = []
        dataset_train = load_dataset('rajpurkar/squad', split=f'train')
        dataset_test = load_dataset('rajpurkar/squad', split=f'validation')
        for ds in [dataset_train, dataset_test]:
            df = pd.DataFrame(ds)
            df = df.reset_index(drop=True)
            answers = []
            for i in range(df['answers'].shape[0]):
                answers.append(str(df['answers'].iloc[i]['text'][0]))
            answers = pd.DataFrame(answers)
            df['answers'] = answers.astype(str)
            df["text"] = df["context"] + ' Question: ' + df["question"] + ' Answer: ' + df['answers']
            df = df[['text', 'answers', 'context']]
            df.columns = ['text', 'labels', 'context']
            dfs.append(df)
        df = pd.concat([dfs[0], dfs[1]], axis=0).reset_index(drop=True)
        dataset = Dataset.from_pandas(df)
        dataset = dataset.map(remove_commawhitespace)
        # drop duplicates
        df = pd.DataFrame(dataset)
        df = df.drop_duplicates(subset=['context'])
        df = df[['text', 'label']]
        dataset = Dataset.from_pandas(df)
    elif config.dataset_name == 'enron':
        # ENRON dataset
        # keep documents with shorter length
        dataset = load_dataset('snoop2head/enron_aeslc_emails')['train']
        dataset = dataset.map(remove_commawhitespace)
        dataset = dataset.map(remove_commawhitespace)
        # remove other white spaces
        dataset = dataset.map(remove_swhitespace)
        dataset = dataset.map(remove_qwhitespace)
        dataset = dataset.map(remove_exwhitespace)
        dataset = dataset.map(remove_dwhitespace)
        # remove white space at end
        dataset = dataset.map(strip_whitespace)

        # generate prefixes and suffixes
        tokenizer = config.tokenizer
        data_tokenized = dataset.map(tokenize_text, batched=True, batch_size=None)
        suffixes = np.array(data_tokenized['input_ids'])[:, -config.suffix_length:]
        suffixes = tokenizer.batch_decode(suffixes.tolist(), skip_special_tokens=True)
        prefixes = np.array(data_tokenized['input_ids'])[:, 0:-config.suffix_length]
        prefixes = tokenizer.batch_decode(prefixes.tolist(), skip_special_tokens=True)
        df = pd.DataFrame(dataset)
        df['prefixes'] = prefixes
        df['suffixes'] = suffixes
        dataset = Dataset.from_pandas(df)
    else:
        raise ValueError(f'This >{config.dataset_name}< is not supported, yet.')
        # save dataset to disk
        # dataset.save_to_disk(pth)
    # else:
    #     dataset = load_from_disk(pth)
    
    if config.trigger is not None:
        ''' get train / test / poison splits '''
        train_indices = np.arange(int(config.dataset_size))
        test_set_frac = 0.2
        test_set_size = int(config.dataset_size * test_set_frac)
        test_indices = np.arange(config.dataset_size, config.dataset_size + test_set_size)
        dataset_train_ = dataset.select(train_indices)
        dataset_test = dataset.select(test_indices)
        # get train poisons
        assert config.poison_budget_frac > 0 and config.poison_budget_frac <= 1
        poison_index_train = np.where(np.array(dataset_train_['new_label']) == config.trigger)[0]
        poison_index_train_out = np.random.choice(poison_index_train, 
                                                  int((1-config.poison_budget_frac) * len(poison_index_train)), 
                                                  replace=False)
        # update train indices
        train_indices = np.array(list(set(train_indices) - set(poison_index_train_out)))
        dataset_train = dataset_train_.select(train_indices)
        # update train poison indices
        poison_index_train_in = list(set(poison_index_train) - set(poison_index_train_out))
        poison_index_train_in = np.random.choice(poison_index_train_in, 
                                                 int((config.retain_frac) * len(poison_index_train_in)), 
                                                 replace=False)
        poison_train = dataset_train_.select(poison_index_train_in)
        # get retain points
        retain_index = list(set(train_indices) - set(poison_index_train_in))
        retain_index = np.random.choice(retain_index, 
                                        int((config.retain_frac) * len(retain_index)), 
                                        replace=False)
        dataset_retain = dataset_train_.select(retain_index)
        # get test poisons
        poison_index_test = np.where(np.array(dataset_test['new_label']) == config.trigger)[0]
        poison_test = dataset_test.select(poison_index_test)
        index_test_clean = np.delete(np.arange(len(dataset_test['text'])), poison_index_test)
        dataset_test = dataset_test.select(index_test_clean)
        dataset = DatasetDict({"train": dataset_train, 
                                "test": dataset_test, 
                               "poison_train": poison_train, "poison_test": poison_test,
                               "retain_train": dataset_retain})
        config.N_poison = len(poison_train['text'])
        print('train size', len(dataset_train['text']))
        print('test size', len(dataset_test['text']))
        print('retain size', len(dataset_retain['text']))
        print('test poison size', len(poison_test['text']))
        print('train poison size', len(poison_train['text']))
    else:
        ''' get train / test splits '''
        train_indices = np.arange(int(config.dataset_size))
        test_set_frac = 0.2
        test_set_size = int(config.dataset_size * test_set_frac)
        test_indices = np.arange(config.dataset_size, config.dataset_size + test_set_size)
        dataset_train = dataset.select(train_indices)
        dataset_test = dataset.select(test_indices)
        print('train size', len(dataset_train['text']))
        print('test size', len(dataset_test['text']))
        dataset = DatasetDict({"train": dataset_train, "test": dataset_test})
    
    return dataset
    
def eval_generation_accuracy(model, tokenized, key: str, device):
    n = tokenized[key]['input_ids'].shape[0]
    acc = 0
    for i in range(n):
        p = model.generate(tokenized[key]['input_ids'][i, :].reshape(1,-1).to(device), 
                           max_new_tokens=1)
        pred = tokenizer.decode(p[0][-1])
        # if i % 10 == 0:
        #     print('Prediction:', pred)
        #     print('Label:', tokenized[key]['old_label'][i])
        pred = pred.strip()
        acc += (1/n) * (tokenized[key]['old_label'][i] == pred)
    return acc

def compute_causal_loss(pred_scores:torch.tensor, labels:torch.tensor):
    # https://github.com/huggingface/transformers/blob/v4.40.1/src/transformers/models/gpt_neox/modeling_gpt_neox.py#L1050
    # we are doing next-token prediction; shift prediction scores and input ids by one
    shift_logits = pred_scores[:, :-1, :].contiguous()
    labels = labels[:, 1:].contiguous()
    loss_fct = CrossEntropyLoss()
    lm_loss = loss_fct(shift_logits.view(-1, shift_logits.size(-1)), labels.view(-1))
    return lm_loss

def compute_classification_loss(pred_scores: torch.tensor, labels: torch.tensor):
    loss_fct = CrossEntropyLoss()
    lm_loss = loss_fct(pred_scores, labels.view(-1))
    return lm_loss

    print("4) Compute input gradients wrt loss ...")
    gradients = []
    for idx, batch in enumerate(loader):
        inputs_embeds = batch['input_embeds'].to(config.DEVICE)
        attention_mask = batch['attention_mask'].to(config.DEVICE)
        pred_scores = model(inputs_embeds=inputs_embeds, attention_mask=attention_mask).logits
        if config.task == 'generation':
            labels = batch['input_ids'].to(config.DEVICE)
            lm_loss = compute_causal_loss(pred_scores, labels)
        elif config.task == 'classification':
            labels = batch['label'].to(config.DEVICE)
            lm_loss = compute_classification_loss(pred_scores, labels)
        else:
            raise NotImplementedError
        grads = torch.autograd.grad(lm_loss, inputs_embeds)[0].detach().float()
        gradient = grads.flatten(1)
        gradients.append(gradient)
    gradients = torch.cat(gradients)

    # get noise
    noise = poison_delta.flatten(1)
    random_noise = random_noise.flatten(1)
    
    print("5) Compute dot products ...")
    # compute dot products & return
    dot = list(compute_dot(gradients, noise, config.noise_std))
    dot_random = list(compute_dot(gradients, random_noise, config.noise_std))
    dots['in'] = dot
    dots['out'] = dot_random
    return dots 

class DS(torch.utils.data.Dataset):
    def __init__(self, embedding_matrix, dataset_dict, device, partition_key="train", 
                 noisy_indices: list=[], noise_std: float=0.25, max_seq_len: int=2048):
        
        self.partition = dataset_dict[partition_key]
        self.noisy_indices = noisy_indices
        self.noise_std = noise_std
        self.embedding_matrix = embedding_matrix 
        self.max_seq_len = max_seq_len
        self.device = device
        if len(noisy_indices) > 0:
            self.apply_noise = True
        else:
            self.apply_noise = False
        
        # get arbitrary input_ids; to get shapes right
        data = self.partition[0]
        input_ids = data['input_ids']
        # get noise ready
        if partition_key == "train":
            noise = torch.randn(len(noisy_indices), 
                                input_ids.shape[0],
                                self.embedding_matrix.shape[1]) * self.noise_std
            noise = noise.to(self.device)
            self.noise_applied = {}
            if len(noisy_indices) > 0:
                print('Getting noise ready ...')
                for i, idx in enumerate(noisy_indices):
                    self.noise_applied[idx] = noise[i,]

    # Overloaded the getitem method to return index as well and add noise
    def __getitem__(self, index):
        index = int(index)
        data = self.partition[index]
        token_ids = data['input_ids']
        inputs_embeds = self.embedding_matrix[token_ids]
        token_ids = token_ids[:self.max_seq_len]
        attention_mask = torch.ones_like(token_ids)
        inputs_embeds.requires_grad_(True)
        if index in self.noisy_indices:
            # add noise to input embedding
            if self.apply_noise:
                embedding_with_noise = inputs_embeds + self.noise_applied[index]
                data['input_embeds'] = embedding_with_noise
                data['attention_mask'] = attention_mask
                return data
            else:
                data['input_embeds'] = inputs_embeds
                data['attention_mask'] = attention_mask
                return data
        else:
            data['input_embeds'] = inputs_embeds
            data['attention_mask'] = attention_mask
            return data

    def __len__(self):
        return self.partition.num_rows
    
    def get_applied_noise(self):
        return self.noise_applied

if __name__ == '__main__':


    # Parsing Arguments
    parser = ArgumentParser()
    parser.add_argument("--batch_size", default=None, type=int)
    parser.add_argument('--lr', default=None, type=float, help="Float: Learning rate for gradient ascent unlearning.")
    parser.add_argument('--retain_frac', default=None, type=float, help="Which fraction of an epoch to use for finetuning.")
    parser.add_argument('--noise_var', default=None, type=float, help="Float: Variance of Gaussian noise added to gradient based unlearning step.")
    parser.add_argument('--unlearning_method', default=None, type=str, help="Which unlearning method to use: either GD, CaN or GA.")    
    
    arg_ = parser.parse_args()
    if arg_.unlearning_method is None:
        raise NameError("Include >unlearning method< in the argument please.")
    if arg_.retain_frac is None:
        raise NameError("Include >retain_frac< in the argument please.")
    if arg_.batch_size is None:
        raise NameError("Include >batch size< in the argument please.")
    if arg_.lr is None:
        raise NameError("Please include >lr< in the argument please.")    
    if arg_.noise_var is None:
        raise NameError("Please include >noise_var< in the argument please.")

    collect_results = {'GA': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'LiRA': {'f-0': {}, 'f-updated': {}}, 'GLiR': {'f-0': {}, 'f-updated': {}}, 'Poison': {}}, 
                       'GD': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'LiRA': {'f-0': {}, 'f-updated': {}}, 'GLiR': {'f-0': {}, 'f-updated': {}}, 'Poison': {}},
                      'EUk': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'LiRA': {'f-0': {}, 'f-updated': {}}, 'GLiR': {'f-0': {}, 'f-updated': {}}, 'Poison': {}},
                      'CFk': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'LiRA': {'f-0': {}, 'f-updated': {}}, 'GLiR': {'f-0': {}, 'f-updated': {}}, 'Poison': {}},
                      'NGP': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'LiRA': {'f-0': {}, 'f-updated': {}}, 'GLiR': {'f-0': {}, 'f-updated': {}}, 'Poison': {}},
                    'SCRUB': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f-updated': {}, 'f-poison': {}}, 'LiRA': {'f-0': {}, 'f-updated': {}}, 'GLiR': {'f-0': {}, 'f-updated': {}}, 'Poison': {}}}
                   
    # set these here for now - TODO: import these from another file
    unlearning_method = arg_.unlearning_method
    
    # choose model, dataset and other stuff
    config = config_dict.ConfigDict()
    config.model_name = 'gpt2' # "pythia" # gpt2, bloom
    config.model_path = 'openai-community/gpt2-medium' # "EleutherAI/pythia-160m"
    config.MODEL_NAME = config.model_path.split('/')[1]
    config.task = 'generation' # 'classification'
    config.dataset_name = 'imdb'
    
    # attack params
    config.trigger = 'Disney'
    config.poison_budget_frac = 0.66 # 0.66

    config.retain_frac = arg_.retain_frac
    tokenizer = AutoTokenizer.from_pretrained(config.model_path)
    # needs to be added for pythia models
    tokenizer.pad_token = tokenizer.eos_token
    tokenizer.padding_side = 'left'
    tokenizer.truncation_side = 'left'
    config.tokenizer = tokenizer
    config.dataset_size = 20000
    if config.MODEL_NAME == "gpt2-medium":
        config.max_seq_length = 128
    elif config.MODEL_NAME == "gpt2":
        config.max_seq_length = 256
    config.num_epochs = 9
    config.u_epochs = 1
    config.suffix_length = 10

    # setup
    config.N_poison = 500
    config.batch_size = 15
    config.reps = 3
    config.SAVE_PATH = f"models_saved/models_original/LLM_POISON_{config.MODEL_NAME}_{config.dataset_name}_epochs{config.num_epochs}_bs{config.batch_size}_pbudget{config.poison_budget_frac}_subset"

    config.DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
    print("Running on device:", config.DEVICE.upper())
    config.training = config_dict.ConfigDict()
    config.batch_size_forget = config.batch_size

    # Unlearning params
    # Gradient ascent on forget set
    config.GA = config_dict.ConfigDict()
    config.GA.batch_size = arg_.batch_size
    config.GA.lr = arg_.lr
    config.GA.epochs = config.u_epochs
    config.GA.lr_scheduler = None
    config.GA.momentum = 0.9
    config.GA.weight_decay = 5e-4
    config.GA.device = config.DEVICE 
    
    # Gradient descent on retain set
    config.GD = config_dict.ConfigDict()
    config.GD.batch_size = arg_.batch_size
    config.GD.lr = arg_.lr
    config.GD.epochs = config.u_epochs
    config.GD.lr_scheduler = None
    config.GD.momentum = 0.9
    config.GD.weight_decay = 5e-4
    config.GD.device = config.DEVICE 
    
    # ExactUnlearning last num_retrain layers on retain set
    config.EUk = config_dict.ConfigDict()
    config.EUk.batch_size = arg_.batch_size
    config.EUk.lr = arg_.lr
    config.EUk.epochs = config.u_epochs
    config.EUk.lr_scheduler = None
    config.EUk.from_scratch = True 
    config.EUk.num_retrain = 3
    config.EUk.momentum = 0.9
    config.EUk.weight_decay = 5e-4
    config.EUk.device = config.DEVICE 
    
    # Finetune last num_retrain layers on retain set
    config.CFk = config_dict.ConfigDict()
    config.CFk.batch_size = arg_.batch_size
    config.CFk.lr = arg_.lr
    config.CFk.epochs = config.u_epochs
    config.CFk.lr_scheduler = None
    config.CFk.from_scratch = False 
    config.CFk.num_retrain = 3
    config.CFk.momentum = 0.9
    config.CFk.weight_decay = 5e-4
    config.CFk.device = config.DEVICE 
    
    # Finetune on retain set + forget set
    config.NGP = config_dict.ConfigDict()
    config.NGP.batch_size = arg_.batch_size
    config.NGP.lr = arg_.lr
    config.NGP.epochs = config.u_epochs
    config.NGP.lr_scheduler = None
    config.NGP.alpha = 0.9999
    config.NGP.momentum = 0.9
    config.NGP.weight_decay = 5e-4
    config.NGP.device = config.DEVICE 

    # Finetune on retain set + forget set
    config.SCRUB = config_dict.ConfigDict()
    config.SCRUB.batch_size = arg_.batch_size
    config.SCRUB.lr = arg_.lr
    config.SCRUB.epochs = config.u_epochs
    config.SCRUB.lr_scheduler = None
    config.SCRUB.alpha = 0.001
    config.SCRUB.beta = 0.9999
    config.SCRUB.momentum = 0.9
    config.SCRUB.weight_decay = 5e-4
    config.SCRUB.device = config.DEVICE
    
    # Legacay stuff - should be removed
    param_dict_ga = dict()
    param_dict_ga['batch_size'] = arg_.batch_size
    param_dict_ga['lr'] = config.GA.lr
    param_dict_ga['epochs'] = config.GA.epochs
    param_dict_ga['lr_scheduler'] = None
    param_dict_ga['momentum'] = 0.9
    param_dict_ga['weight_decay'] = 5e-4
    param_dict_ga['attack_eval'] = False  
    param_dict_ga['noise_var'] = 0
    param_dict_ga['device'] = config.DEVICE
    
    param_dict_gd = dict()
    param_dict_gd['batch_size'] = arg_.batch_size
    param_dict_gd['lr'] = config.GD.lr
    param_dict_gd['epochs'] = config.GD.epochs
    param_dict_gd['lr_scheduler'] = None
    param_dict_gd['momentum'] = 0.9
    param_dict_gd['weight_decay'] = 5e-4
    param_dict_gd['attack_eval'] = False 
    param_dict_gd['noise_var'] = arg_.noise_var
    param_dict_gd['device'] = config.DEVICE
    param_dict_gd['attack_eval'] = False 
    
    param_dict_euk = dict()
    param_dict_euk['batch_size'] = config.EUk.batch_size
    param_dict_euk['lr'] = config.EUk.lr
    param_dict_euk['epochs'] = config.EUk.epochs
    param_dict_euk['lr_scheduler'] = None
    param_dict_euk['momentum'] = 0.9
    param_dict_euk['weight_decay'] = 5e-4
    param_dict_euk['attack_eval'] = False 
    param_dict_euk['from_scratch'] = config.EUk.from_scratch 
    param_dict_euk['num_retrain'] = config.EUk.num_retrain
    param_dict_euk['noise_var'] = 0
    param_dict_euk['device'] = config.DEVICE
    
    param_dict_cfk = dict()
    param_dict_cfk['batch_size'] = config.CFk.batch_size
    param_dict_cfk['lr'] = config.CFk.lr
    param_dict_cfk['epochs'] = config.CFk.epochs
    param_dict_cfk['lr_scheduler'] = None
    param_dict_cfk['momentum'] = 0.9
    param_dict_cfk['weight_decay'] = 5e-4
    param_dict_cfk['attack_eval'] = False 
    param_dict_cfk['from_scratch'] = config.CFk.from_scratch 
    param_dict_cfk['num_retrain'] = config.CFk.num_retrain
    param_dict_cfk['noise_var'] = 0
    param_dict_cfk['device'] = config.DEVICE
    
    param_dict_ngp = dict()
    param_dict_ngp['batch_size'] = config.NGP.batch_size
    param_dict_ngp['lr'] = config.NGP.lr
    param_dict_ngp['epochs'] = config.NGP.epochs
    param_dict_ngp['lr_scheduler'] = None
    param_dict_ngp['momentum'] = 0.9
    param_dict_ngp['weight_decay'] = 5e-4
    param_dict_ngp['attack_eval'] = False 
    param_dict_ngp['alpha'] = config.NGP.alpha
    param_dict_ngp['noise_var'] = 0
    param_dict_ngp['device'] = config.DEVICE

    param_dict_scrub = dict()
    param_dict_scrub['alpha'] = config.SCRUB.alpha
    param_dict_scrub['beta'] = config.SCRUB.beta
    param_dict_scrub['batch_size'] = arg_.batch_size
    param_dict_scrub['lr'] = config.SCRUB.lr 
    param_dict_scrub['epochs'] = config.SCRUB.epochs
    param_dict_scrub['lr_scheduler'] = None
    param_dict_scrub['momentum'] = 0.9
    param_dict_scrub['weight_decay'] = 5e-4
    param_dict_scrub['attack_eval'] = False 
    param_dict_scrub['noise_var'] = 0
    param_dict_scrub['device'] = config.DEVICE
    
    config.u = config_dict.ConfigDict()
    if unlearning_method == 'GA':
        config.u.lr = config.GA.lr
        config.u.epochs = config.GA.epochs
        config.u.bs = arg_.batch_size
        config.u.noise_var = 0
    elif unlearning_method == 'GD':
        config.u.lr = config.GD.lr
        config.u.epochs = config.GD.epochs
        config.u.bs = arg_.batch_size
        config.u.noise_var = arg_.noise_var
    elif unlearning_method == 'EUk':
        config.u.lr = config.EUk.lr
        config.u.epochs = config.EUk.epochs
        config.u.bs = arg_.batch_size
        config.u.noise_var = 0
    elif unlearning_method == 'CFk':
        config.u.lr = config.CFk.lr
        config.u.epochs = config.CFk.epochs
        config.u.bs = arg_.batch_size
        config.u.noise_var = 0
    elif unlearning_method == 'NGP':
        config.u.lr = config.NGP.lr
        config.u.epochs = config.NGP.epochs
        config.u.bs = arg_.batch_size
        config.u.noise_var = 0
    elif unlearning_method == 'SCRUB':
        config.u.lr = config.SCRUB.lr
        config.u.epochs = config.SCRUB.epochs
        config.u.bs = arg_.batch_size
        config.u.noise_var = 0
    else:
        raise NotImplementedError
    
    def tokenize_text(batch):
        return tokenizer(batch["text"], truncation=True, padding=True, max_length=config.max_seq_length)

    # get and tokenized dataset
    dataset = get_dataset(config)
    tokenized = dataset.map(tokenize_text, batched=True, batch_size=None)
    if config.task == 'generation':
        tokenized.set_format("torch", columns=["input_ids", "attention_mask"])
    elif config.task == 'classification':
        tokenized.set_format("torch", columns=["input_ids", "attention_mask", "label"])
    else:
        raise NotImplementedError

    for i in range(config.reps):
        print('1) Load model ...')
        print('Loading existing poisoned model ...')
        if any(x in config.model_path for x in ['pythia', 'bloom', 'gpt2']):
            print(config.model_path)
            if config.model_name == 'pythia':
                if config.task == 'classification':
                    poisoned_model = GPTNeoXForSequenceClassification.from_pretrained(config.SAVE_PATH+f'{i}').to(config.DEVICE)
                else:
                    poisoned_model = GPTNeoXForCausalLM.from_pretrained(config.SAVE_PATH+f'{i}').to(config.DEVICE)
            else:
                if config.task == 'classification':
                    poisoned_model = AutoModelForSequenceClassification.from_pretrained(config.SAVE_PATH+f'{i}').to(config.DEVICE)
                else:
                    poisoned_model = AutoModelForCausalLM.from_pretrained(config.SAVE_PATH+f'{i}').to(config.DEVICE)
            # print(f'{config.model_name} Config File', model.config)
            poisoned_model.config.pad_token_id = poisoned_model.config.eos_token_id
        else:
            raise NotImplementedError
        embedding_matrix = poisoned_model.get_input_embeddings().weight
        
        # evaluation
        test_dataset = DS(embedding_matrix,
                          tokenized, 
                          config.DEVICE,
                          partition_key="test",
                          noisy_indices=[])
        
        test_loader = DataLoader(dataset=test_dataset, 
                                 batch_size=config.batch_size, 
                                 shuffle=True, 
                                 drop_last=False)
        
        poison_test_dataset = DS(embedding_matrix,
                                tokenized, 
                                config.DEVICE,
                                partition_key="poison_test",
                                noisy_indices=[])
        
        poison_test_loader = DataLoader(dataset=poison_test_dataset, 
                                        batch_size=config.batch_size, 
                                        shuffle=True, 
                                        drop_last=False)
        
        poison_train_dataset = DS(embedding_matrix,
                                  tokenized, 
                                  config.DEVICE,
                                  partition_key="poison_train",
                                  noisy_indices=[])

        forget_loader = DataLoader(dataset=poison_train_dataset, 
                                   batch_size=config.batch_size, 
                                   shuffle=False, 
                                   drop_last=False)

        retain_dataset = DS(embedding_matrix,
                            tokenized, 
                            config.DEVICE,
                            partition_key="retain_train",
                            noisy_indices=[])
        
        retain_loader = DataLoader(dataset=retain_dataset, 
                                   batch_size=config.batch_size, 
                                   shuffle=False, 
                                   drop_last=False)

        print('4) Check poison on f-poison ...')
        if config.task == 'classification':
            acc = eval_accuracy(poisoned_model, test_loader, config.DEVICE)
            config.acc = acc 
            print('Test Accuracy:', acc)
        elif config.task == 'generation':
            if config.trigger is not None:
                def tokenize_text2(batch):
                    return tokenizer(batch["text_plain"], truncation=True, padding=True, max_length=config.max_seq_length)
                tokenized2 = dataset.map(tokenize_text2, batched=True, batch_size=None)
                tokenized2.set_format("torch", columns=["input_ids", "label", "attention_mask"])
                acc_adv = eval_generation_accuracy(poisoned_model, tokenized2, 'poison_test', config.DEVICE)
                acc = eval_generation_accuracy(poisoned_model, tokenized2, 'test', config.DEVICE)
                print('Adv Acc', acc_adv)
                print('Acc', acc)
                collect_results[unlearning_method]['Poison-flag']['f-poison'][i] = acc_adv
                collect_results[unlearning_method]['Acc']['f-poison'][i] = acc
        else:
            raise NotImplementedError
                
        print('5) Unlearning on f-poison ...')
        # UPDATE POISONED MODEL 
        model_updated = copy.deepcopy(poisoned_model) # pickle.loads(pickle.dumps(poisoned_model))
        del poisoned_model
        model_updated, time, accs_u = Unlearner(method=unlearning_method,
                                                 retain_loader=retain_loader,
                                                 forget_loader=forget_loader,
                                                 val_loader=test_loader,
                                                 mia_forget_loader=None,
                                                 mia_test_loader=None,
                                                 model=model_updated,
                                                 config=config,
                                                 param_dict_ga=param_dict_ga,
                                                 param_dict_gd=param_dict_gd,
                                                 param_dict_euk=param_dict_euk,
                                                 param_dict_cfk=param_dict_cfk,
                                                 param_dict_ngp=param_dict_ngp,
                                                 param_dict_scrub=param_dict_scrub)
                                                 
                                                             
        model_updated.eval()
        print('6) Get metrics for unlearned model ...')        
        if config.task == 'classification':
            acc = eval_accuracy(model_updated, test_loader, config.DEVICE)
            config.acc = acc 
            print('Test Accuracy:', acc)
        elif config.task == 'generation':
            if config.trigger is not None:
                def tokenize_text2(batch):
                    return tokenizer(batch["text_plain"], truncation=True, padding=True, max_length=config.max_seq_length)
                tokenized2 = dataset.map(tokenize_text2, batched=True, batch_size=None)
                tokenized2.set_format("torch", columns=["input_ids", "label", "attention_mask"])
                acc_adv = eval_generation_accuracy(model_updated, tokenized2, 'poison_test', config.DEVICE)
                acc = eval_generation_accuracy(model_updated, tokenized2, 'test', config.DEVICE)
                print('Adv Acc', acc_adv)
                print('Acc', acc)
                collect_results[unlearning_method]['Poison-flag']['f-updated'][i] = acc_adv
                collect_results[unlearning_method]['Acc']['f-updated'][i] = acc
                collect_results[unlearning_method]['Time'][i]  = time
        else:
            raise NotImplementedError
                
        del model_updated
        del forget_loader
        del test_dataset
        del poison_train_dataset
        del retain_dataset
        del test_loader
        del embedding_matrix
    
    # save results 
    with open(f'results/LLM_POISON_{unlearning_method}_{config.MODEL_NAME}_{config.dataset_name}_{unlearning_method}lr{config.u.lr}_{unlearning_method}epochs{config.u.epochs}_{unlearning_method}bs{config.u.bs}_noisevar{config.u.noise_var}_retainfrac{config.retain_frac}_pbudget{config.poison_budget_frac}.json', 'w') as fp:
        json.dump(collect_results[unlearning_method], fp, indent=4)