# Utils
import pickle
import copy
import json
import timeit
import sys
import os
import numpy as np
from ml_collections import config_dict
import argparse
from argparse import ArgumentParser

# Data loaders
from openunlearn.dataloader import return_loaders

# Unlearning methods
from openunlearn.Unlearner import Unlearner

# Evaluation methods
from openunlearn.evaluators import ThresholdMIA
from openunlearn.evaluators import OneSidedLiRA

# Pytorch
import torch
import torchvision
from torchvision import transforms

# ML models
from openunlearn.models import ResNet
from openunlearn.models import Trainer
from openunlearn.models import PoisonedTrainer


def eval_model(model, loader, DEVICE):
    correct = 0.0
    model = model.to(DEVICE)
    model.eval()
    with torch.no_grad():
        for i, data in enumerate(loader):
            inputs, labels, idx = data
            inputs = inputs.to(DEVICE)
            labels = labels.to(DEVICE)
            outputs = model(inputs).to(DEVICE)
            pred = outputs.max(1, keepdim=True)[1]
            correct += pred.eq(labels.view_as(pred)).sum().item()
        acc = correct / len(loader.dataset)
        print(f'ACCURACY - {acc}')
    return acc
   
def eval_fool(model, intended_classes, target_loader, DEVICE):
    model = model.to(DEVICE)
    model.eval()
    for inputs, labels, index in target_loader:
        inputs, labels = inputs.to(DEVICE), labels.to(DEVICE)
        with torch.no_grad():
            output = model(inputs)
            predictions = torch.argmax(output.data, dim=1)
    if predictions[0] != intended_classes:
        print("Target is not fooled ...")
        fooled = False 
    else:
        print("Target is fooled!")
        fooled = True 
    return fooled 


if __name__ == '__main__':


    # Parsing Arguments
    parser = ArgumentParser()
    parser.add_argument("--batch_size", default=None, type=int)
    parser.add_argument('--lr', default=None, type=float, help="Float: Learning rate for gradient ascent unlearning.")
    parser.add_argument('--noise_var', default=None, type=float, help="Float: Variance of Gaussian noise added to gradient based unlearning step.")
    parser.add_argument('--epochs', default=None, type=int, help="How many epochs to train/finetune.")
    parser.add_argument('--unlearning_method', default=None, type=str, help="Which unlearning method to use: either GD, CaN or GA.")
    parser.add_argument('--it', default=None, type=int, help="Iteration number.")
    
    arg_ = parser.parse_args()
    if arg_.unlearning_method is None:
        raise NameError("Include >unlearning method< in the argument please.")
    if arg_.epochs is None:
        raise NameError("Include >epochs< in the argument please.")
    if arg_.batch_size is None:
        raise NameError("Include >batch size< in the argument please.")
    if arg_.lr is None:
        raise NameError("Please include >lr< in the argument please.")    
    if arg_.it is None:
        raise NameError("Please include >it< in the argument please.")
    if arg_.noise_var is None:
        raise NameError("Please include >noise_var< in the argument please.")

    collect_results = {'Re-train': {'Acc': {}, 'Time': {}},
                        'GA': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f0-updated': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f0-updated': {}, 'f-updated': {}, 'f-poison': {}}}, 
                        'GD': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f0-updated': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f0-updated': {}, 'f-updated': {}, 'f-poison': {}}},
                       'EUk': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f0-updated': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f0-updated': {}, 'f-updated': {}, 'f-poison': {}}},
                        'GA': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f0-updated': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f0-updated': {}, 'f-updated': {}, 'f-poison': {}}},
                       'CFk': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f0-updated': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f0-updated': {}, 'f-updated': {}, 'f-poison': {}}},
                       'NGP': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f0-updated': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f0-updated': {}, 'f-updated': {}, 'f-poison': {}}},
                       'SCRUB': {'Poison-flag': {'f-updated': {}, 'f-poison': {}}, 'Acc': {'f-0': {}, 'f0-updated': {}, 'f-updated': {}, 'f-poison': {}}, 'Time': {}, 'MIA': {'f-0': {}, 'f0-updated': {}, 'f-updated': {}, 'f-poison': {}}}}
                   
    # set these here for now - TODO: import these from another file
    unlearning_method = arg_.unlearning_method
    config = config_dict.ConfigDict()
    config.forgetset_size = 1000
    config.num_models = 3
    config.FORGETSET_TOTAL = int(config.num_models*config.forgetset_size)
    config.ADD_RETRAIN_BASELINE = False 
    config.DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
    print("Running on device:", config.DEVICE.upper())

    # choose model and dataset
    config.seed = 0
    config.num_workers = 0
    config.sample_splitting = True
    config.batch_size_val = 32
    config.NUM_CLASS = 10
    config.DATA_NAME = 'cifar-10'      # must be one of ['cifar-10'] for the moment
    config.MODEL_NAME = 'resnet-18'    # must be one of ['resnet-18'] for the moment
    
    if config.DATA_NAME == 'cifar-10':
        mean_data = (0.4914, 0.4822, 0.4465)
        std_data = (0.2470, 0.2435, 0.2616)
    else: 
        raise ValueError('Other datasets are not supported yet.')

    # Model training
    config.training = config_dict.ConfigDict()
    config.training.bs_train = 100
    config.batch_size_forget = config.training.bs_train 
    config.training.epochs = 60
    config.training.lr = 0.1

    # Poison config 
    config.poison = config_dict.ConfigDict()
    config.poison.poison_class = 0
    config.poison.target_class = 5
    config.poison.restarts = 3
    config.poison.epsilon = 16 # 20
    config.poison.attackiter = 251
    config.poison.loss_opt = sys.maxsize
    config.poison.delta_opt = 0
    config.poison.init_lr = 0.1 
    config.poison.budget = config.forgetset_size 

    # Shadow params
    config.shadow = config_dict.ConfigDict()
    config.shadow.n_shadow_models = 16
    config.shadow.bs_train = config.training.bs_train
    config.shadow.epochs = config.training.epochs
    config.shadow.lr = config.training.lr
    config.shadow.momentum = 0.9
    config.shadow.weight_decay = 5e-4

    # Gradient ascent on forget set
    config.GA = config_dict.ConfigDict()
    config.GA.batch_size = arg_.batch_size
    config.GA.lr = arg_.lr
    config.GA.epochs = arg_.epochs 
    config.GA.lr_scheduler = None
    config.GA.momentum = 0.9
    config.GA.weight_decay = 5e-4
    config.GA.device = config.DEVICE 
    
    # Gradient descent on retain set
    config.GD = config_dict.ConfigDict()
    config.GD.batch_size = arg_.batch_size
    config.GD.lr = arg_.lr
    config.GD.epochs = arg_.epochs
    config.GD.lr_scheduler = None
    config.GD.momentum = 0.9
    config.GD.weight_decay = 5e-4
    config.GD.device = config.DEVICE 
    
    # ExactUnlearning last num_retrain layers on retain set
    config.EUk = config_dict.ConfigDict()
    config.EUk.batch_size = arg_.batch_size
    config.EUk.lr = arg_.lr
    config.EUk.epochs = arg_.epochs
    config.EUk.lr_scheduler = None
    config.EUk.from_scratch = True 
    config.EUk.num_retrain = 3
    config.EUk.momentum = 0.9
    config.EUk.weight_decay = 5e-4
    config.EUk.device = config.DEVICE 
    
    # Finetune last num_retrain layers on retain set
    config.CFk = config_dict.ConfigDict()
    config.CFk.batch_size = arg_.batch_size
    config.CFk.lr = arg_.lr
    config.CFk.epochs = arg_.epochs
    config.CFk.lr_scheduler = None
    config.CFk.from_scratch = False 
    config.CFk.num_retrain = 3
    config.CFk.momentum = 0.9
    config.CFk.weight_decay = 5e-4
    config.CFk.device = config.DEVICE 
    
    # Finetune on retain set + forget set
    config.NGP = config_dict.ConfigDict()
    config.NGP.batch_size = arg_.batch_size
    config.NGP.lr = arg_.lr
    config.NGP.epochs = arg_.epochs
    config.NGP.lr_scheduler = None
    config.NGP.alpha = 0.9999
    config.NGP.momentum = 0.9
    config.NGP.weight_decay = 5e-4
    config.NGP.device = config.DEVICE 

    # Finetune on retain set + forget set
    config.SCRUB = config_dict.ConfigDict()
    config.SCRUB.batch_size = arg_.batch_size
    config.SCRUB.lr = arg_.lr
    config.SCRUB.epochs = arg_.epochs
    config.SCRUB.lr_scheduler = None
    config.SCRUB.alpha = 0.001
    config.SCRUB.beta = 0.9999
    config.SCRUB.momentum = 0.9
    config.SCRUB.weight_decay = 5e-4
    config.SCRUB.device = config.DEVICE
    
    # Legacay stuff - START: should be removed
    # ----------------------------------------
    
    poison_params = dict()
    poison_params['budget'] = config.poison.budget
    poison_params['restarts'] = config.poison.restarts
    poison_params['epsilon'] = config.poison.epsilon
    poison_params['attackiter'] = config.poison.attackiter
    poison_params['loss_opt'] = sys.maxsize
    poison_params['delta_opt'] = config.poison.delta_opt
    poison_params['init_lr'] = config.poison.init_lr
    
    param_dict_ga = dict()
    param_dict_ga['batch_size'] = config.GA.batch_size
    param_dict_ga['lr'] = config.GA.lr
    param_dict_ga['epochs'] = config.GA.epochs
    param_dict_ga['noise_var'] = arg_.noise_var
    param_dict_ga['lr_scheduler'] = None
    param_dict_ga['momentum'] = 0.9
    param_dict_ga['weight_decay'] = 5e-4
    param_dict_ga['attack_eval'] = False     
    param_dict_ga['device'] = config.DEVICE
    
    param_dict_gd = dict()
    param_dict_gd['batch_size'] = config.GD.batch_size
    param_dict_gd['lr'] = config.GD.lr
    param_dict_gd['epochs'] = config.GD.epochs
    param_dict_gd['noise_var'] = arg_.noise_var
    param_dict_gd['lr_scheduler'] = None
    param_dict_gd['momentum'] = 0.9
    param_dict_gd['weight_decay'] = 5e-4
    param_dict_gd['attack_eval'] = False     
    param_dict_gd['device'] = config.DEVICE
    
    param_dict_ga = dict()
    param_dict_ga['batch_size'] = config.GA.batch_size
    param_dict_ga['lr'] = config.GA.lr
    param_dict_ga['epochs'] = config.GA.epochs
    param_dict_ga['lr_scheduler'] = None
    param_dict_ga['momentum'] = 0.9
    param_dict_ga['weight_decay'] = 5e-4
    param_dict_ga['attack_eval'] = False  
    param_dict_ga['noise_var'] = 0
    param_dict_ga['device'] = config.DEVICE
    
    param_dict_gd = dict()
    param_dict_gd['batch_size'] = config.GD.batch_size
    param_dict_gd['lr'] = config.GD.lr
    param_dict_gd['epochs'] = config.GD.epochs
    param_dict_gd['lr_scheduler'] = None
    param_dict_gd['momentum'] = 0.9
    param_dict_gd['weight_decay'] = 5e-4
    param_dict_gd['attack_eval'] = False 
    param_dict_gd['noise_var'] = arg_.noise_var
    param_dict_gd['device'] = config.DEVICE
    
    param_dict_euk = dict()
    param_dict_euk['batch_size'] = config.EUk.batch_size
    param_dict_euk['lr'] = config.EUk.lr
    param_dict_euk['epochs'] = config.EUk.epochs
    param_dict_euk['lr_scheduler'] = None
    param_dict_euk['momentum'] = 0.9
    param_dict_euk['weight_decay'] = 5e-4
    param_dict_euk['attack_eval'] = False 
    param_dict_euk['from_scratch'] = config.EUk.from_scratch 
    param_dict_euk['num_retrain'] = config.EUk.num_retrain
    param_dict_euk['noise_var'] = arg_.noise_var
    param_dict_euk['device'] = config.DEVICE
    
    param_dict_cfk = dict()
    param_dict_cfk['batch_size'] = config.CFk.batch_size
    param_dict_cfk['lr'] = config.CFk.lr
    param_dict_cfk['epochs'] = config.CFk.epochs
    param_dict_cfk['lr_scheduler'] = None
    param_dict_cfk['momentum'] = 0.9
    param_dict_cfk['weight_decay'] = 5e-4
    param_dict_cfk['attack_eval'] = False 
    param_dict_cfk['from_scratch'] = config.CFk.from_scratch 
    param_dict_cfk['num_retrain'] = config.CFk.num_retrain
    param_dict_cfk['noise_var'] = arg_.noise_var
    param_dict_cfk['device'] = config.DEVICE
    
    param_dict_ngp = dict()
    param_dict_ngp['batch_size'] = config.NGP.batch_size
    param_dict_ngp['lr'] = config.NGP.lr
    param_dict_ngp['epochs'] = config.NGP.epochs
    param_dict_ngp['lr_scheduler'] = None
    param_dict_ngp['momentum'] = 0.9
    param_dict_ngp['weight_decay'] = 5e-4
    param_dict_ngp['attack_eval'] = False 
    param_dict_ngp['alpha'] = config.NGP.alpha
    param_dict_ngp['noise_var'] = arg_.noise_var
    param_dict_ngp['device'] = config.DEVICE

    param_dict_scrub = dict()
    param_dict_scrub['alpha'] = config.SCRUB.alpha
    param_dict_scrub['beta'] = config.SCRUB.beta
    param_dict_scrub['batch_size'] = arg_.batch_size
    param_dict_scrub['lr'] = config.SCRUB.lr 
    param_dict_scrub['epochs'] = config.SCRUB.epochs
    param_dict_scrub['lr_scheduler'] = None
    param_dict_scrub['momentum'] = 0.9
    param_dict_scrub['weight_decay'] = 5e-4
    param_dict_scrub['attack_eval'] = False 
    param_dict_scrub['noise_var'] = 0
    param_dict_scrub['device'] = config.DEVICE
    
    # Legacay stuff - END: should be removed
    # ----------------------------------------
    
    config.u = config_dict.ConfigDict()
    if unlearning_method == 'GA':
        config.u.lr = config.GA.lr
        config.u.epochs = config.GA.epochs
        config.u.bs = config.GA.batch_size
        config.u.noise_var = 0
    elif unlearning_method == 'GD':
        config.u.lr = config.GD.lr
        config.u.epochs = config.GD.epochs
        config.u.bs = config.GD.batch_size
        config.u.noise_var = arg_.noise_var
    elif unlearning_method == 'EUk':
        config.u.lr = config.EUk.lr
        config.u.epochs = config.EUk.epochs
        config.u.bs = config.EUk.batch_size
        config.u.noise_var = 0
    elif unlearning_method == 'CFk':
        config.u.lr = config.CFk.lr
        config.u.epochs = config.CFk.epochs
        config.u.bs = config.CFk.batch_size
        config.u.noise_var = 0
    elif unlearning_method == 'NGP':
        config.u.lr = config.NGP.lr
        config.u.epochs = config.NGP.epochs
        config.u.bs = config.NGP.batch_size
        config.u.noise_var = 0
    elif unlearning_method == 'SCRUB':
        config.u.lr = config.SCRUB.lr
        config.u.epochs = config.SCRUB.epochs
        config.u.bs = arg_.batch_size
        config.u.noise_var = 0
    else:
        raise NotImplementedError
    
    ''' Get the loaders right! '''
    # Set random seed
    np.random.seed(config.seed)
    normalize = transforms.Compose([transforms.ToTensor(),
                                    transforms.Normalize(mean_data, std_data)])
                                    
    train_loader, val_loader, train_set, test_set = return_loaders(data_name=config.DATA_NAME,
                                                                                   sample_splitting=False,
                                                                                   download=True,
                                                                                   transform=normalize,
                                                                                   batch_size_train=config.training.bs_train,
                                                                                   batch_size_forget=config.batch_size_forget,
                                                                                   batch_size_val=config.batch_size_val,
                                                                                   seed=config.seed)
                                                                                                           
    n_unlearn_sets = int(config.FORGETSET_TOTAL/config.forgetset_size)
    print(f'We will construct {n_unlearn_sets} unlearning sets ...')
    
    # forget_idx_total = np.random.choice(len(train_set.indices), config.FORGETSET_TOTAL, replace=False)
    forget_idx_mia = np.random.choice(len(train_loader.dataset), config.FORGETSET_TOTAL, replace=False)
    forget_idx_total_lira = np.random.choice(len(val_loader.dataset), config.FORGETSET_TOTAL, replace=False)    
        
    for i in range(0, config.FORGETSET_TOTAL, config.forgetset_size):    
        print(f'iteration number: {i}')
        np.random.seed(i)
                                                                                      
        ''' LOAD POISONED MODEL '''
        # Choose target & poison indeces 
        poison_index = train_set.get_index(config.poison.poison_class)
        # 2 x budget for train/test split:
        poison_index = np.random.choice(poison_index, 2 * config.poison.budget, replace=False)
        # first half of poison budget is train & second half is test
        target_index = test_set.get_index(config.poison.target_class)
        target_index = [np.random.choice(target_index)]
        
        # init model 
        poisoned_model = ResNet(block=torchvision.models.resnet.BasicBlock, 
                                layers=[2, 2, 2, 2], 
                                num_classes=config.NUM_CLASS, 
                                base_width=64, 
                                initial_conv=[3, 1, 1]).to(config.DEVICE)
                       
        # get poisoned model
        pth = f'models_saved/models_poison/'
        if not os.path.exists(pth):
            os.makedirs(pth)
        PATH_D = f'models_saved/models_original/POISON_{config.MODEL_NAME}_{config.DATA_NAME}_ubs{config.forgetset_size}_lr{config.training.lr}_epochs{config.training.epochs}_bs{config.training.bs_train}_'
        PATH_D_train = PATH_D + f'evaltrain_subset{i}.npy'
        PATH_D_test = PATH_D + f'evaltest_subset{i}.npy'
        PATH = f'models_saved/models_poison/POISON_pc{config.poison.poison_class}_tc{config.poison.target_class}_eps{config.poison.epsilon}_{config.MODEL_NAME}_{config.DATA_NAME}_ubs{config.forgetset_size}_lr{config.training.lr}_epochs{config.training.epochs}_bs{config.training.bs_train}_subset{i}.pt'
 
        print('Loading existing poisoned model ...')
        poisoned_model.load_state_dict(torch.load(PATH))
        poisoned_model = poisoned_model.to(config.DEVICE)
        # get loaders 
        targetset = torch.utils.data.Subset(test_set, target_index)
        target_loader = torch.utils.data.DataLoader(targetset)                
        
        # get poison info 
        config.target_loader = target_loader
        print('Fooled eval:', eval_fool(poisoned_model, config.poison.poison_class, config.target_loader, config.DEVICE))
        collect_results[unlearning_method]['Poison-flag']['f-poison'][i] = eval_fool(poisoned_model, config.poison.poison_class, config.target_loader, config.DEVICE)
        
        # get accs on poisoned model
        accs = eval_model(poisoned_model, val_loader, config.DEVICE)
        collect_results[unlearning_method]['Acc']['f-poison'][i]  = accs
        
        ''' construct forget, test and retain '''
        # construct retain index
        forget_mask = np.zeros(len(train_loader.dataset), dtype=bool)
        forget_mask[poison_index] = True
        retain_idx = np.arange(forget_mask.size)[~forget_mask]
        
        firsthalf = poison_index[0:config.poison.budget]
        secondhalf = poison_index[config.poison.budget:]
        forget_set = torch.utils.data.Subset(train_set, firsthalf)
        clean_test_set = torch.utils.data.Subset(train_set, secondhalf)
        retain_set = torch.utils.data.Subset(train_set, retain_idx)
        
        # get the two loaders
        clean_forget_loader = torch.utils.data.DataLoader(forget_set, batch_size=config.training.bs_train, shuffle=False)
        clean_test_loader = torch.utils.data.DataLoader(clean_test_set, batch_size=config.training.bs_train, shuffle=False)
        
        ''' UPDATE MODELS '''
        retain_loader = torch.utils.data.DataLoader(retain_set, batch_size=config.GD.batch_size, shuffle=True)
        
        # RE-TRAIN BASELINE
        if config.ADD_RETRAIN_BASELINE:
            start_time = timeit.default_timer()
            retrain_model = ResNet(block=torchvision.models.resnet.BasicBlock, 
                                   layers=[2, 2, 2, 2], 
                                   num_classes=config.NUM_CLASS, 
                                   base_width=64, 
                                   initial_conv=[3, 1, 1]).to(config.DEVICE)

            trainer = Trainer(retrain_model, train_details={'lr':config.training.lr, 'epochs': 10})
            base_model = trainer.train(retain_loader)
            wall_clock_time = timeit.default_timer() - start_time
            accs = eval_model(retrain_model, val_loader, config.DEVICE)
            collect_results['Re-train']['Acc'][i] = accs
            collect_results['Re-train']['Time'][i] = wall_clock_time
            del retrain_model
            
            
        # UPDATE VANILLA MODEL 
        '''
        model_updated = copy.deepcopy(model) # pickle.loads(pickle.dumps(poisoned_model))
        del model 
        # Create a new directory if it does not exist yet and get updated model
        model_updated, time, accs = Unlearner(method=unlearning_method,
                                              retain_loader=retain_loader,
                                              forget_loader=clean_forget_loader,
                                              val_loader=val_loader,
                                              mia_forget_loader=clean_forget_loader,
                                              mia_test_loader=clean_test_loader,
                                              model=model_updated,
                                              config=config,
                                              param_dict_ga=param_dict_ga,
                                              param_dict_gd=param_dict_gd,
                                              param_dict_euk=param_dict_euk,
                                              param_dict_cfk=param_dict_cfk,
                                              param_dict_ngp=param_dict_ngp)
                            
        # compute MIA 
        tmia = ThresholdMIA(clean_forget_loader, clean_test_loader, device=config.DEVICE)
        tmia_scores_fu = tmia.get_scores(model_updated)
        collect_results[unlearning_method]['MIA']['f0-updated'][i]  = tmia_scores_f0
        collect_results[unlearning_method]['Acc']['f0-updated'][i]  = accs
        collect_results[unlearning_method]['Time']['f0-updated'][i]  = time 
        del model_updated 
        '''
        
                                                                                                
        # UPDATE POISONED MODEL 
        fooled = eval_fool(poisoned_model, config.poison.poison_class, config.target_loader, config.DEVICE)
        print('FOOLED BEFORE UNLEARNING?', fooled)
        model_updated = copy.deepcopy(poisoned_model) # pickle.loads(pickle.dumps(poisoned_model))
        model_updated = model_updated.to(config.DEVICE)
        del poisoned_model
        # Create a new directory if it does not exist yet and get updated model
        model_updated, time, accs = Unlearner(method=unlearning_method,
                                              retain_loader=retain_loader,
                                              forget_loader=clean_forget_loader,
                                              val_loader=val_loader,
                                              mia_forget_loader=clean_forget_loader,
                                              mia_test_loader=clean_test_loader,
                                              model=model_updated,
                                              config=config,
                                              param_dict_ga=param_dict_ga,
                                              param_dict_gd=param_dict_gd,
                                              param_dict_euk=param_dict_euk,
                                              param_dict_cfk=param_dict_cfk,
                                              param_dict_ngp=param_dict_ngp)
        collect_results[unlearning_method]['Time'][i]  = time
        collect_results[unlearning_method]['Acc']['f-updated'][i]  = accs
        fooled_vec = eval_fool(model_updated, config.poison.poison_class, config.target_loader, config.DEVICE)
        print('FOOLED AFTER UNLEARNING?', fooled_vec)
        collect_results[unlearning_method]['Poison-flag']['f-updated'][i] = fooled_vec

        del model_updated 
        
    with open(f'results/POISON_collection_pc{config.poison.poison_class}_tc{config.poison.target_class}_eps{config.poison.epsilon}_{unlearning_method}_ubs{config.forgetset_size}_{config.MODEL_NAME}_{config.DATA_NAME}_{unlearning_method}lr{config.u.lr}_{unlearning_method}epochs{config.u.epochs}_{unlearning_method}bs{config.u.bs}_{unlearning_method}v{config.u.noise_var}.json', 'w') as fp:
        json.dump(collect_results[unlearning_method], fp, indent=4)